/**
 * \file mgui/appmain.h
 * \brief Definitions for MGUI::APP_MAIN and MGUI::APP_BASE classes
 *
 * \if NODOC
 * $Id: appmain.h_v 1.33 2011/11/22 13:56:13 mju Exp $
 *
 * $Log: appmain.h_v $
 * Revision 1.33  2011/11/22 13:56:13  mju
 * Remove mfc stuff.
 *
 * Revision 1.29  2009/04/24 17:59:44  bcolombe
 * Added AllowedProducts to APP_JOB and make v_InitInstance protected
 *
 * Revision 1.28  2009/04/24 16:41:01  bcolombe
 * Added APP_JOB
 *
 * Revision 1.27  2009/04/23 17:48:26  bcolombe
 * Added v_GetSessionLogPath
 * Gets a specific path to use for the session log
 * Empty if want to use default
 *
 * Revision 1.25  2009/01/08 17:07:47  mju
 * Add support for queued delegates.
 *
 * Revision 1.24  2008/11/14 16:06:08  mju
 * Add getCmdParmPath.
 *
 * Revision 1.21  2006/08/23 22:43:38  dwilliss
 * Added icon support
 *
 * Revision 1.5  2003/07/29 16:33:30  mju
 * Add SetLegacyAllowShellResize.
 *
 * Revision 1.1  2003/04/23 21:59:55  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MGUI_APPMAIN_H
#define	INC_MGUI_APPMAIN_H

#ifndef INC_MI32_ICONID_H
	#include <mi32/iconid.h>
#endif

#if !defined(_XtIntrinsic_h)
	#include <X11/Intrinsic.h>
#endif

#ifndef	INC_MI32_MILIST_H
	#include <mi32/milist.h>
#endif

#ifndef	INC_MI32_DELEGATE_H
	#include <mi32/delegate.h>
#endif

#ifndef	INC_MI32_CRITSEC_H
	#include <mi32/critsec.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
struct McbAnyCallbackStruct;
class FILEPATH;
namespace MGUI {
	class APP_MAIN;
	class TIMER;
	}
#endif


namespace MGUI {

//=====================================================================================================================

//! Base class for application process instance.
//! A class derived from this must be constructed by a class derived from APP_MAIN,
//! either in v_InitInstance or in its constructor.  Failure to follow this
//! convention will result in an exception during application startup or shutdown.
//! Note that this is handled automatically for applications based on MGUI::APP_MULTI.
class APP_BASE {
	public:

		enum LEGACYINIT {
			LEGACYINIT_Default =					0x00,
			LEGACYINIT_NoSetPosition =			0x01,	//!< Don't automatically set position
			LEGACYINIT_NoSetTitle =				0x02,	//!< Don't automatically set title
			LEGACYINIT_NoRealize =				0x04,	//!< Don't 'realize' toplevel application shell after InitInstance
			LEGACYINIT_NoAutoSetExit =			0x08,	//!< Don't do MxSetExitProcess() when toplevel shell 'closed' via Window Manager
			LEGACYINIT_LockShellSize =			0x10,	//!< Lock toplevel shell size after realizing
			LEGACYINIT_NoDestroyTopOnExit =	0x20	//!< Don't destroy TopLevel shell on exit as application destroys it
			};

		//! Get top level parent window for dialogs
		//! Intended for use by MGUI::SHELL, not usually called directly.
		static MDLGPARENT GetTopDlgParent ();
	
		//! Get number of command-line parameters.
		inline int GetCmdParmCount () const;

		//! Get command-line parameter as string for specified index.
		inline const char* GetCmdParm (
			int idx								//!< Command parameter index
			) const;

		//! Get command-line parameter as FILEPATH.
		//! @return true if have parameter, false if not, does not check for path existence.
		bool GetCmdParmPath (
			int idx,
			FILEPATH& pathret
			) const;

		//! Get the default ICONID for MGUI::SHELLs
		ICONID GetDefaultIconID (
			) const { return (v_GetDefaultIconID()); }

	protected:

		//! Constructor.
		APP_BASE (
			UINT32 AllowedProduct1 = 0,		//!< Product code for license checking
			UINT32 AllowedProduct2 = 0,
			UINT32 AllowedProduct3 = 0,
			UINT32 AllowedProduct4 = 0
			);

		//! Destructor
		virtual ~APP_BASE ();

		//! Perform legacy initialization for processes using TopLevel shell (X only).
		//! Replaces MxInitProc(), should call from within v_InitInstance().
		//! @return Unrealized TopLevel shell.
		//! @deprecated Use MGUI::TOPSHELL as base for main application window.
		DEPRECATED Widget MxInitLegacy (LEGACYINIT flags = LEGACYINIT_Default);

		//! Set to allow resize of legacy toplevel shell.
		//! Must be done in derived class constructor.
		//! Required for some applications which need to dynamically grow the legacy toplevel shell.
		void SetLegacyAllowShellResize (
			) { m_LegacyAllowShellResize = true; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		LEGACYINIT m_LegacyInitFlags;
		bool m_LegacyAllowShellResize;
		MDLGPARENT m_TopParent;
		UINT32 m_ProductCode;
		UINT32 m_AllowedProducts[4];
		MILIST<DELEGATE_VOID_NOPARMS> m_QueuedDelegates;
		MGUI::TIMER *m_pQueueTimer;
		MITHREAD::CRITICALSECTION m_QueueCritSec;

		void AddQueuedDelegate (DELEGATE_VOID_NOPARMS delegate);
		void EnableQueueDelegates (bool enable);
		UINT32 DetermineProductCode ()
			{ return (v_DetermineProductCode()); }
		void ExitInstance ();
		const char* GetProcessName () const
			{ return (v_GetProcessName()); }
		bool InitInstance ();
		bool IsProductAllowed (UINT32) const;
		void OnQueueTimer ();
		void RemoveQueuedDelegate (DELEGATE_VOID_NOPARMS delegate);

		friend class MGUI::APP_MAIN;

		APP_BASE (const APP_BASE&);
		APP_BASE& operator= (const APP_BASE&);

		#endif // GENERATING_DOXYGEN_OUTPUT

		// Overridables.

		//! Override to determine product code for license check.
		//! Default implementation determines product code based on command
		//! line parameters and initialization settings.  Not normally overridden.
		virtual UINT32 v_DetermineProductCode () const;

		//! Override to perform resource cleanup prior to application exit.
		virtual void v_ExitInstance ();

		//! Override to change the default ICONID for the icon in the upper-left
		//! corner of the title bar of all MGUI::SHELLs.
		//! If not overridden, the default is to not set the icons, although this
		//! may change to become a TNTmips icon
		virtual ICONID v_GetDefaultIconID () const;

		//! Must be overridden to return 'process name' for use as 'INI group' and 'registry subkey'.
		//! Should return character string having no whitespace.
		virtual const char* v_GetProcessName () const = 0;

		//! Get specific path used for session log for this application.  Returns false if none.
		virtual void v_GetSessionLogPath (FILEPATH& path);

		//! Must be overridden to perform initialization of this application.
		//! Initialization normally includes creation of primary user interface
		//! windows prior to entering main message handling loop.
		//! @return 'true' if successful, 'false' if not (will exit if false).
		virtual bool v_InitInstance () = 0;

		//! Called at startup just prior to MsysAppInitialize.
		//! @return true to continue, false to terminate application.
		virtual bool v_PreInitInstance ();

	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(APP_BASE::LEGACYINIT)
#endif

//=====================================================================================================================

//! Base class for main application instance.
//! A class must be derived from APP_MAIN and a single instance must be
//! declared at file (static global) scope in the main application module.
//! Only one instance of this class may be defined per executable.
//! See also: APP_BASE.
class APP_MAIN {
	public:

		//! Add queued delegate, which will be called from main thread in main loop.
		//! If delegate is already in queue it will not be added again.
		//! Can be called from worker thread.
		static void AddQueuedDelegate (
			DELEGATE_VOID_NOPARMS delegate
			);

		//! Enabled queued delegates - may only be called from main thread.
		static void EnableQueueDelegates (
			bool enable = true
			);

		//! Terminate the event loop and exit application normally.
		static void Exit () { s_pMainInstance->DoExit(); }

		//! Get reference to the single APP_BASE instance for this executable.
		static APP_BASE* GetBaseInstance ()
			{ return (s_pBaseInstance); }

		//! Get the default icon for MGUI::SHELLs
		static ICONID GetDefaultIconID ();

		//! Get reference to the single APP_MAIN instance for this executable.
		static APP_MAIN* GetMainInstance ()
			{ return (s_pMainInstance); }

		//! Process single event if any is pending and return.
		//! This function should be used in long-running code where no status information is being
		//! displayed but event processing is desired.  This function should be used sparingly.
		static void ProcessPendingEvent ();

		//! Process all pending events (if any) and return.
		//! This function can be used to force display to "catch" up with changes before
		//! returning to user control.  This function should be used sparingly.
		static void ProcessPendingEvents ();

		//! Remove queued delegate.
		static void RemoveQueuedDelegate (
			DELEGATE_VOID_NOPARMS delegate
			);

		//! Constructor.
		APP_MAIN ();

		//! Destructor.
		virtual ~APP_MAIN () = 0;

		//! Get number of command-line parameters.
		int GetCmdParmCount () const
			{ return (m_CmdParmCount); }

		//! Get command-line parameter for specified index.
		const char* GetCmdParm (
			int idx								//!< Command parameter index
			) const { return ((idx < 0 || idx >= m_CmdParmCount) ? NULL : m_CmdParmArray[idx]); }

	#ifdef WIN32
		//! Get Windows application instance handle (Windows only).
		void *GetInstanceHandle () const
			{ return (m_hInstance); }
	#endif

		//! Remove command-line parameter at specified index.
		//! Note, idx must be > 0 as first entry cannot be removed.
		void RemoveCmdParm (
			int idx
			);

		#ifndef GENERATING_DOXYGEN_OUTPUT
	#if defined(WIN32)
		void DoWinMain (void *hInstance, const char *cmdline);
	#else
		// These should be private, but then main() can't access them.  I tried
		// friend int main(int, char**), but it didn't help.
		void Run();
		void SetCmdParms (const char**, int);
	#endif
		#endif // GENERATING_DOXYGEN_OUTPUT

	private:

		#ifndef GENERATING_DOXYGEN_OUTPUT

		static APP_MAIN *s_pMainInstance;
		static APP_BASE *s_pBaseInstance;
	#ifdef WIN32
		void *m_hInstance;
		char *m_ModNameStr;
		char *m_CmdLineStr;
	#endif

		void DoExit ();
		void ExitInstance ();
		bool InitInstance ();
		static void OnExitXServer (McbAnyCallbackStruct*, void*);
		static void OnTerminate ();

	#if defined(WIN32)
		void Run ();
	#endif

	#if defined(WIN32)
		void SetCmdLine (const char*);
	#endif
		char **m_CmdParmArray;
		int m_CmdParmCount;

		friend class MGUI::APP_BASE;

		APP_MAIN (const APP_MAIN&);
		APP_MAIN& operator= (const APP_MAIN&);

		#endif // GENERATING_DOXYGEN_OUTPUT

		// Overridables.

		//! Called just prior to application exit.
		//! May be overridden to clean up allocated instance resource.
		virtual void v_ExitInstance ();

		//! Called at application startup.
		//! Normally overridden if multiple 'processes' are built into a single executable.
		//! In this case the command line parameters may be processed to determine which
		//! process to invoke and call a function or method to create a static instance of
		//! a derived APP_BASE class for that process.
		virtual bool v_InitInstance ();

	};

//=====================================================================================================================

//! Base class for main application when run under job processing.
class APP_JOB : public APP_BASE {
	public:
		APP_JOB (
			UINT32 AllowedProduct1 = 0,		//!< Product code for license checking
			UINT32 AllowedProduct2 = 0,
			UINT32 AllowedProduct3 = 0,
			UINT32 AllowedProduct4 = 0
			);
		~APP_JOB () {}

	protected:
		virtual bool v_InitInstance ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		void ExecuteJob (FILEPATH path);

		virtual ERRVALUE v_RunJob(const FILEPATH& path);

		// Overrides from APP_BASE.
		virtual void v_GetSessionLogPath (FILEPATH& path);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

#ifndef GENERATING_DOXYGEN_OUTPUT

//! Get number of command-line parameters.
inline int APP_BASE::GetCmdParmCount (
	) const {
	APP_MAIN *pMain = APP_MAIN::GetMainInstance();
	return ((pMain == 0) ? 0 : pMain->GetCmdParmCount());
	}

//! Get command-line parameter for specified index.
inline const char* APP_BASE::GetCmdParm (
	int idx								//!< Command parameter index
	) const {
	APP_MAIN *pMain = APP_MAIN::GetMainInstance();
	return ((pMain == 0) ? 0 : pMain->GetCmdParm(idx));
	}

#endif // GENERATING_DOXYGEN_OUTPUT

//=====================================================================================================================

}	// End namespace MGUI

#endif	// INC_MGUI_APPMAIN_H
