/**
 * \file mgui/canvas.h
 * \brief MGUI "canvas" control definitions.
 *
 * \if NODOC
 * $Id: canvas.h_v 1.49 2012/03/22 20:53:19 mju Exp $
 *
 * $Log: canvas.h_v $
 * Revision 1.49  2012/03/22 20:53:19  mju
 * Add updateFromBuffer.
 *
 * Revision 1.48  2012/03/13 17:08:00  mju
 * Add getDrawingDevice with parm to allow stereo or not.
 *
 * Revision 1.47  2012/01/10 15:57:56  mju
 * Add stereoDrawCursor.
 *
 * Revision 1.46  2012/01/09 21:40:30  mju
 * Add getCursor and cursorSetVisible.
 *
 * Revision 1.45  2011/12/30 16:58:45  mju
 * Add getStyle.
 *
 * Revision 1.44  2011/12/30 16:01:51  mju
 * Remove invalidate methods as replaced by restoreFromBuffer.
 *
 * Revision 1.43  2011/12/29 18:28:39  mju
 * Remove enums that are no longer used.
 *
 * Revision 1.42  2011/12/28 22:34:09  mju
 * Add separate buffering for pixmap vs memory.
 *
 * Revision 1.41  2011/12/08 22:22:28  mju
 * Add getDrawingContext with stereoview.
 *
 * Revision 1.40  2011/12/08 18:49:40  mju
 * Make internals private to implementation so can extend without rebuild.
 *
 * Revision 1.39  2011/12/07 18:45:12  mju
 * Change how stereo enabled/disabled so retains device hardware description.
 *
 * Revision 1.38  2011/12/05 22:00:01  mju
 * Add stereodevice member.
 *
 * Revision 1.37  2011/10/28 15:11:08  mju
 * Change to use dual-buffering for stereo instead of double width.
 *
 * Revision 1.36  2011/10/24 20:53:03  mju
 * Add side-by-side stereo buffer and method to update into regular buffer.
 *
 * Revision 1.34  2010/10/27 15:59:00  mju
 * Add getHeight/Width/PointerLocation.
 *
 * Revision 1.33  2010/10/25 14:35:49  mju
 * Add "hasmore" to v_onPointerMove and define DELEGATE_ONPOINTERMOVE to correspond.
 *
 * Revision 1.32  2010/02/11 21:24:36  mju
 * Add style_TraverseToOnPointerEnter.
 *
 * Revision 1.31  2009/10/09 22:00:01  mju
 * Add double-click delegate.
 *
 * Revision 1.29  2009/09/08 20:43:32  mju
 * Add style_PixmapBuffer.
 *
 * Revision 1.25  2008/02/08 20:40:49  mju
 * Add style_noResizeDelay.
 *
 * Revision 1.1  2002/10/11 20:32:18  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MGUI_CANVAS_H
#define	INC_MGUI_CANVAS_H

#ifndef INC_MGUI_CTRL_H
	#include <mgui/ctrl.h>
#endif

#ifndef INC_MGUI_KEYCODE_H
	#include <mgui/keycode.h>
#endif

#ifndef INC_MGUI_CURSORID_H
	#include <mgui/cursorid.h>
#endif

#ifndef INC_MGUI_POINTERBUTTON_H
	#include <mgui/pointerbutton.h>
#endif

#ifndef	INC_MI32_STEREOVIEW_H
	#include <mi32/stereoview.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
class STEREODEVICE;
namespace MGD {
	class DEVICE;
	class CONTEXT;
	}
#endif


namespace MGUI {

//===================================================================================================================
//!	Generic 'canvas' control supporting drawing and simple user input.
class CANVAS : public CTRL {
	public:

		enum DEVICETYPE {
			DEVICETYPE_Screen,			//!< Screen device associated with canvas window
			DEVICETYPE_Pixmap,			//!< Pixmap buffer, set as background for window
			DEVICETYPE_Memory,			//!< Memory buffer device, size will match canvas window
			DEVICETYPE_Stereo,			//!< Stereo buffer, will have 2 frame buffers
			};

		enum STYLE {
			STYLE_Default =						0x00,
			STYLE_UsePixmap =						0x01,		//!< Use "background pixmap" for off-screen buffering
			STYLE_UseMemory =						0x02,		//!< Use "memory" for off-screen buffering
			STYLE_NoBuffer =						0x04,		//!< Don't use buffer unless specifically requested via STYLE_UsePixmap or STYLE_UseMemory
			STYLE_NoDelayResize =				0x08,		//!< Don't use delayed notification of resize
			STYLE_NotifyExpose =					0x10,		//!< Notify via v_OnPaint when area exposed for redrawing (automatic if using delegate)
			STYLE_TraverseToOnPointerEnter =	0x20,		//!< Traverse (make input for keypress) when pointer enters canvas
			};

		typedef fastdelegate::FastDelegate<bool(MGUI::POINT,MGUI::KEYCODE,MGUI::KEYSTATE)> DELEGATE_ONKEY;
		typedef fastdelegate::FastDelegate<bool(MGUI::POINT,MGUI::POINTERBUTTON,MGUI::KEYSTATE)> DELEGATE_ONPOINTER;
		typedef fastdelegate::FastDelegate<bool(MGUI::POINT,MGUI::POINTERBUTTON,MGUI::KEYSTATE,bool)> DELEGATE_ONPOINTERMOVE;
		typedef fastdelegate::FastDelegate<void(MGD::CONTEXT*)> DELEGATE_ONPAINT;
		typedef fastdelegate::FastDelegate<void(const LRECT2D&)> DELEGATE_ONUPDATERECT;
		typedef fastdelegate::FastDelegate<void(INT32)> DELEGATE_ONWHEEL;

		//! Constructor.
		CANVAS (
			STYLE style = STYLE_Default
			);

		//! Destructor.
		virtual ~CANVAS ();

		//! Add delegate to call when pointer button pressed.
		//! The delegate gets passed the same parameters as v_OnPointerButtonDown().
		//! Delegate must return 'true' if it processed the message, false to continue processing.
		void AddDelegateOnPointerButtonDown (
			POINTERBUTTON buttons,				//!< Button(s) for which to call delegate
			DELEGATE_ONPOINTER delegate		//!< Delegate to call
			);

		//! Add delegate to call when pointer button released.
		//! The delegate gets passed the same parameters as v_OnPointerButtonUp().
		//! Delegate must return 'true' if it processed the message, false to continue processing.
		void AddDelegateOnPointerButtonUp (
			POINTERBUTTON buttons,				//!< Button(s) for which to call delegate
			DELEGATE_ONPOINTER delegate		//!< Delegate to call
			);

		//! Create canvas control
		void Create (
			LAYOUT_PANE_BASE& ParentPane,		//!< Parent pane
			int width,								//!< Desired width in pixels
			int height,								//!< Desired height in pixels
			LAYOUT_SIZEALIGN sizealign,		//!< Sizing and alignment options
			CANVAS::STYLE style = STYLE_Default
			);

		//! Set cursor visibility.
		void CursorSetVisible (
			bool visible							//!< Whether visible or not
			);

		//! Erase canvas to specified color.
		void Erase (
			const COLOR& color					//!< Color to erase to
			);

		//! Get current cursor ID.
		CURSORID GetCursor () const;

		//! Get suitable parent for dialog launched from canvas.
		MDLGPARENT GetDlgParent () const;

		//! Get shared drawing context.
		//! @return Drawing Context, do not delete as is internal to CANVAS.
		//! Typically used for graphic tools and overlays.
		//! Copy of pointer should not be retained as it may become invalid based on user action or use of StereoDisable/Enable.
		MGD::CONTEXT* GetDrawingContext (
			DEVICETYPE DeviceType,
			STEREOVIEW StereoView = STEREOVIEW_None	//!< Stereo view, ignored unless stereo enabled
			) const;

		//! Get drawing device for buffer or direct display.
		//! Do not delete the returned DEVICE as it is internal to the CANVAS.
		MGD::DEVICE* GetDrawingDevice (
			DEVICETYPE DeviceType = DEVICETYPE_Memory		//!< Drawing device type
			);

		//! Get drawing device for buffer or direct display with control over whether to allow stereo.
		//! Do not delete the returned DEVICE as it is internal to the CANVAS.
		MGD::DEVICE* GetDrawingDevice (
			DEVICETYPE DeviceType,								//!< Drawing device type
			bool AllowStereo										//!< Whether to allow stereo left/right frames (if supported)
			);

		//! Get height of canvas in pixels.
		INT16 GetHeight () const;

		//! Get current pointer location relative to canvas.
		//! @return 'true' if within canvas extent, 'false' if not.
		//! If returns false, ptCanvas will have coordinates of INT16_MIN if could not determine location at all.
		bool GetPointerLocation (
			MGUI::POINT& ptCanvas				//!< Location relative to canvas returned
			) const;

		//! Get dimensions of canvas.
		void GetSize (
			INT16& width,							//!< Width in pixels returned
			INT16& height							//!< Height in pixels returned
			) const;

		//! Get canvas style flags.
		STYLE GetStyle () const;

		//! Get width of canvas in pixels.
		INT16 GetWidth () const;

		//! Determine if pointer is "captured" by this canvas.
		bool IsPointerCaptured () const;

		//! Pan view by specified amount by copying from background buffer.
		//! Will update screen display.
		void PanFromBuffer (
			INT32 dx,								//!< X offset in pixels
			INT32 dy,								//!< Y offset in pixels
			const COLOR *pColorBG				//!< Color to fill portions where no data available to copy
			);

		//! Release capture on pointer.
		//! Use with care, typically used if buttonpress pops up a menu.
		void ReleaseCapture ();

		//! Resize canvas
		void Resize (
			INT16 width,							//!< New width requested
			INT16 height							//!< New height requested
			);

		//! Restore entire canvas from specific off-screen buffer.
		//! Will generate OnPaint notifications if STYLE_NotifyExpose was specified in constructor or Create.
		void RestoreFromBuffer (
			DEVICETYPE devtype,					//!< Buffer device type
			bool UpdateNow = false				//!< Perform immediate update before returning
			);

		//! Restore portion of canvas from specific off-screen buffer.
		//! Will generate OnPaint notifications if STYLE_NotifyExpose was specified in constructor or Create.
		//! Note that if DEVICETYPE_Stereo is specified the actual area(s) restored may differ from the specified rectangle.
		void RestoreFromBuffer (
			DEVICETYPE devtype,					//!< Buffer device type
			const LRECT2D& rect,					//!< Portion of canvas to restore
			bool UpdateNow = false				//!< Perform immediate update before returning
			);

		//! Set pointer cursor to show while in canvas.
		void SetCursor (
			CURSORID cursorid
			);

		//! Set delegate to call when key "pressed".
		//! Delegate should @return 'true' if key handled, 'false' if not.
		//! The delegate gets passed the same parameters as v_OnKey()
		void SetDelegateOnKey (
			DELEGATE_ONKEY delegate
			);

		//! Set delegate to call when canvas needs to be redrawn.
		//! This delegate is called after the buffer is updated to the screen.  If no overlay and not unbuffered
		//! then this delegate does not need to be set.
		//! The delegate gets passed the same parameters as v_OnPaint()
		void SetDelegateOnPaint (
			DELEGATE_ONPAINT delegate
			);

		//! Set delegate to call for double-click on any button.
		void SetDelegateOnPointerDoubleClick (
			DELEGATE_ONPOINTER delegate
			);

		//! Set delegate to call when pointer enters canvas area.
		void SetDelegateOnPointerEnter (
			DELEGATE_VOID_NOPARMS delegate
			);

		//! Set delegate to call when pointer leaves canvas area.
		void SetDelegateOnPointerLeave (
			DELEGATE_VOID_NOPARMS delegate
			);

		//! Set delegate to call when pointer is moved.
		//! The delegate gets passed the same parameters as v_OnPointerMove()
		void SetDelegateOnPointerMove (
			DELEGATE_ONPOINTERMOVE delegate
			);

		//! Set delegate to call when the user scrolls the pointer wheel.
		//! The delegate is passed an integer parameter which will be positive
		//! if the wheel was scrolled up and negative if scrolled down.
		void SetDelegateOnPointerWheel (
			DELEGATE_ONWHEEL delegate
			);

		//! Set delegate to call when canvas size changes.
		//! Note, this is NOT called repeatedly while the user resizes the window, but instead will be
		//! called after the user releases the pointer or pauses for a nominal period of time (.5 second).
		//! The delegate gets passed the same parameters as v_OnSize()
		void SetDelegateOnSize (
			DELEGATE_ONSIZE delegate
			);

		//! Set delegate to call when Pixmap updated from Memory buffer.
		//! Only applies if both STYLE_UsePixmap and STYLE_UseMemory specified in construction/creation.
		void SetDelegateOnUpdatePixmap (
			DELEGATE_ONUPDATERECT delegate
			);

		//! Disable stereo buffering.
		void StereoDisable ();

		//! Draw stereo cursor using specified left/right points.
		void StereoDrawCursor (
			MGUI::POINT ptCanvasL,						//!< Left canvas coordinate
			MGUI::POINT ptCanvasR,						//!< Right canvas coordinate
			LRECT2D& CursorRectL,						//!< Left rectangle returned
			LRECT2D& CursorRectR							//!< Right rectangle returned
			);

		//! Enable stereo buffering.
		void StereoEnable (
			const STEREODEVICE& StereoDevice			//! Stereo device hardware description
			);

		//! Get stereo device hardware description.
		//! @return pointer to hardware description, NULL if not available (ie, disabled).
		const STEREODEVICE* StereoGetDeviceHW () const;

		//! Determine if stereo buffering is enabled.
		bool StereoIsEnabled () const;

		//! Update buffer/screen from stereo buffer if enabled and available.
		void StereoUpdateFromBuffer ();

		//! Update specified target device buffer from specified source buffer.
		void UpdateFromBuffer (
			DEVICETYPE DevTarget,						//!< Target device type to update to
			DEVICETYPE DevSource,						//!< Source device type to update from
			const LRECT2D *pRectSource = 0			//!< Rectangle in source buffer, NULL for entire extent
			);

	protected:

		//! Called when cursor "capture" is changed (lost).
		//! Default implementation calls the delegate.
		virtual void v_OnCaptureChanged ();

		//! Called when key "pressed".
		//! Override must call base for proper behavior if does not handle key itself.
		//! @return 'true' if key handled, 'false' if not.
		//! Default implementation calls the delegate.
		virtual bool v_OnKey (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			MGUI::KEYCODE keycode,				//!< Processed key code, equivalent to MIUNICODE for alphanumeric keys.
			MGUI::KEYSTATE keystate				//!< Modifier key state
			);

		//! Called when canvas needs to be redrawn, do not call directly except from derived implementation.
		//! Override must call base for proper behavior.
		//! This is called after the buffer is updated to the screen.  If no overlay and not unbuffered
		//! then this method does not need to be overridden.
		//! Default implementation calls the delegate.
		virtual void v_OnPaint (
			MGD::CONTEXT *gc						//!< Drawing context for unbuffered/overlay.
			);

		//! Called when pointer button pressed.
		//! Override must call base for proper behavior if it does not handle the button itself.
		//! Default implementation calls the delegate(s) associated with the button.
		virtual void v_OnPointerButtonDown (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			POINTERBUTTON button,				//!< Button that was pressed
			KEYSTATE keystate						//!< Modifier key state at time of event
			);

		//! Called when pointer button released.
		//! Return 'true' if it processed the message, false to continue processing.
		//! Default implementation calls the delegate(s) associated with the button.
		virtual void v_OnPointerButtonUp (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			POINTERBUTTON button,				//!< Button that was released
			KEYSTATE keystate						//!< Modifier key state at time of original button press
			);

		//! Called when pointer enters canvas area.
		//! Override must call base for proper behavior.
		//! Default implementation calls the delegate.
		virtual void v_OnPointerEnter ();

		//! Called when pointer leaves canvas area.
		//! Override must call base for proper behavior.
		//! Default implementation calls the delegate.
		virtual void v_OnPointerLeave ();

		//! Called when pointer cursor is moved.
		//! Override must call base for proper behavior.
		//! Default implementation calls the delegate.
		virtual void v_OnPointerMove (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			POINTERBUTTON button,				//!< Current button state
			KEYSTATE keystate,					//!< Modifier key state at time of event
			bool hasmore							//!< Indicates whether more "move" events follow with same button/key state
			);

		//! Called when the user scrolls the pointer wheel
		//! Override must call base for proper behavior.
		//! Default implementation calls the delegate.
		virtual void v_OnPointerWheel (
			INT32 direction						//!< positive for up, negative for down
			);

		//! Called when canvas size changes.
		//! Override must call base for proper behavior.
		//! Note, this is NOT called repeatedly while the user resizes the window, but instead will be
		//! called after the user releases the pointer or pauses for a nominal period of time (.5 second).
		//! Default implementation calls the delegate.
		virtual void v_OnSize (
			int width,								//!< New canvas width in pixels
			int height								//!< New canvas height in pixels
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		class PRIV;
		PRIV *m_pPriv;

		CANVAS (const CANVAS&);
		CANVAS& operator= (const CANVAS&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		//! Called when Pixmap updated from Memory buffer.
		//! Only used if both STYLE_UsePixmap and STYLE_UseMemory specified in construction/creation.
		//! Default implementation does nothing.
		virtual void v_OnUpdatePixmap (
			const LRECT2D& rect					//!< Rectangle that was updated
			);

	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(CANVAS::STYLE)
#endif

//===================================================================================================================

}	// End namespace MGUI


#endif	// INC_MGUI_CANVAS_H
