/**
 * \file mgui/editbase.h
 * \brief MGUI CTRL_EDIT_BASE definitions.
 *
 * \if NODOC
 * $Id: editbase.h_v 1.9 2013/01/18 22:39:36 mju Exp $
 *
 * $Log: editbase.h_v $
 * Revision 1.9  2013/01/18 22:39:36  mju
 * Adjust virtual methods to allow override of on onChangeSelection.
 * Remove some legacy unimplemented methods.
 *
 * Revision 1.8  2011/11/22 13:40:10  mju
 * Fix dox dependencies and exclude mfc stuff.
 *
 * Revision 1.7  2008/10/27 13:36:34  bcolombe
 * Added delegates for gain and lose focus
 *
 * Revision 1.6  2008/09/25 15:09:36  mju
 * Add createflags enum and use in createCtrl.
 *
 * Revision 1.5  2008/09/09 21:54:32  mju
 * Fix doxygen warnings.
 *
 * Revision 1.4  2007/03/16 19:20:03  mju
 * Make setWindowTextMFC non-static.
 *
 * Revision 1.3  2007/03/16 17:54:07  mju
 * Add setDisplayNoValue and string to contain.
 *
 * Revision 1.2  2007/01/31 17:05:37  dwilliss
 * Added yet another bool parameter to CreateCtrl
 *
 * Revision 1.1  2007/01/19 15:48:18  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MGUI_EDITBASE_H
#define	INC_MGUI_EDITBASE_H

#ifndef	INC_MGUI_CTRL_H
#include <mgui/ctrl.h>
#endif

#ifndef	INC_MI32_MISTRING_H
#include <mi32/mistring.h>
#endif

namespace MGUI {

//===================================================================================================================
//! Base class for 'edit' controls.
class CTRL_EDIT_BASE : public MGUI::CTRL {
	public:

		//! Flags for CreateCtrl.
		enum CREATEFLAGS {
			CREATEFLAG_None =						0x00000000,
			CREATEFLAG_ReadOnly =				0x00010000,	//!< Disallow editing
			CREATEFLAG_RightAlign =				0x00020000,	//!< Right-align text
			CREATEFLAG_Opaque =					0x00040000,	//!< Mask values (for password-style) input
			CREATEFLAG_UseFixedWidthFont =	0x00080000,	//!< Use fixed-width font if possible
			CREATEFLAG_WordWrap =				0x00100000,	//!< Word-wrap text instead of showing scrollbar (multi-line only)
			CREATEFLAG_WidthInPixels =			0x00200000,	//!< Width specified in pixels
			CREATEFLAG_HeightInPixels =		0x00400000,	//!< Height specified in pixels (multi-line only)
			CREATEFLAG_SizeInPixels =			(CREATEFLAG_WidthInPixels|CREATEFLAG_HeightInPixels),
			CREATEFLAG_MASK =						0xFFFF0000
			};

		//! Constructor.
		CTRL_EDIT_BASE ();

		//! Destructor.
		~CTRL_EDIT_BASE ();

		//! Clear (delete) the current selection without copying to the clipboard
		void ClearSelection ();

		//! Copy the current selection to the clipboard
		void CopySelection ();

		//! Cut the current selection to the clipboard
		void CutSelection ();

		//! Get the current cursor position
		INT32 GetCursorPos () const;

		//! Returns true if contents of the edit-control have been modified by user, false if not.
		//! Use SetModify() to reset.
		const bool GetModify (
			) const { return (m_UserEdited); }

		//! Get the current selection of the edit control.
		//!	If nStartChar == nEndChar, nothing is selected.
		void GetSelection (
			int& nStartChar,		//!< Receives the first character in the current selection
			int& nEndChar			//!< Receives the first non-selected character past the end of the current selection
			) const;

		//! Insert text at the current cursor location
		void Insert (
			const MISTRING& text
			);

		//! Determine if the edit control has a value
		bool HasValue () const;

		//! Paste the clipboard contents at the current location
		void Paste ();

		//! Replaces the current selection in an edit control with new text.
		//! This replaces only a portion of the text.  To replace all the text,
		//! use SetValue().  If there is no current selection, the new text is
		//! inserted at the current cursor location.
		void ReplaceSelection (
			const MISTRING& NewText
			);

		//! Set the color for ANSI text.
		//! This changes the colors used for the ANSI escape sequences
		//!   \<ESC\>[30m - \<ESC\>[37m   (foreground colors)
		//! 	\<ESC\>[40m - \<ESC\>[47m	(background colors)
		//! The index parameter must be in the range of 0 - 7
		//! The default colors are...
		//!
		//!	0 - black
		//!	1 - red
		//!	2 - green
		//!	3 - yellow
		//!	4 - blue
		//!	5 - magenta
		//!	6 - cyan
		//!	7 - white
		void SetANSIColor (
			INT32 indx,
			const COLOR& color
			);

		//! Set delegate to call when user presses \<Enter\> in edit control.
		void SetDelegateOnActivate (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnActivate = delegate; }

		//! Set delegate to call when value is changed after validation.
		//! If the user modifies the text in the edit control will not be called until the control loses "focus".
		void SetDelegateOnChangeValue (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnChangeValue = delegate; }

		//! Set delegate to call when ctrl gains focus
		void SetDelegateOnGainFocus (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnGainFocus = delegate; }

		//! Set delegate to call when ctrl loses focus
		void SetDelegateOnLoseFocus (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnLoseFocus = delegate; }

		//! Set delegate to call when user selects or unselects text.
		//! Call GetSelection to see what is selected.
		void SetDelegateOnSelection (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnChangeSelection = delegate; }

		//! Set delegate to call when user modifies the text in the edit control.
		//! The actual value does not get updated until the control loses focus or GetValue() is called.
		void SetDelegateOnUserEdit (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnUserEdit = delegate; }

		//! Set string to display when no value is available.
		//! Usually used to prompt user to enter a value, default is blank.
		void SetDisplayNoValue (
			const MISTRING& string
			);

		//! Set the maximim text length in characters
		void SetMaxLength (
			int length
			);

		//! Set or clear the "modified" state of the edit control.
		void SetModify (
			bool modified = true
			) { m_UserEdited = modified; }

		//! Set whether value is read-only or not.
		void SetReadOnly (
			bool ReadOnly = true						//!< True if read only, false if not
			);

		//! Select range of characters in string.
		//! If StartChar is -1, any current selection is unselected.
		void SetSelection (
			int StartChar = 0,						//!< Starting character position in string, 0 for beginning
			int EndChar = -1,							//!< Ending character position, -1 for end of string
			bool NoScroll = true						//!< Indicates whether to scroll caret into view
			);

		//! Set set of 'valid' or 'invalid' characters.
		//! This does not alter the current contents, if any.
		void SetValidChars (
			const MIUNICODE* string,					//!< Characters valid or invalid to enter
			bool invert = false						//!< Allow all but specified characters
			);

	protected:

		bool m_UserEdited;
		bool m_HasValue;

		//! Check that all characters in string are in valid set.
		//!	@return 'true' if valid, 'false' if not
		bool CheckValidChars (
			MISTRING& string								//!< String to be checked
			) const;

		void CreateCtrl (
			LAYOUT_PANE_BASE& ParentPane,
			int width,										//!< Width in 'typical' characters
			int height,										//!< Height in text lines, 0 for single-line edit
			CREATEFLAGS flags,
			LAYOUT_SIZEALIGN sizealign
			);

		bool GetEditStr (
			MISTRING& str
			) const;

		//! Send value change notification.
		void NotifyChanged ();

		//! Called internally to update text control from specified string.
		void UpdateCtrl (
			const MISTRING& string
			);

		//! Called internally to update control-specific data from edit control.
		void UpdateValue (bool notify);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		DELEGATE_VOID_NOPARMS m_DelegateOnActivate;
		DELEGATE_VOID_NOPARMS m_DelegateOnChangeValue;
		DELEGATE_VOID_NOPARMS m_DelegateOnUserEdit;
		DELEGATE_VOID_NOPARMS m_DelegateOnChangeSelection;
		DELEGATE_VOID_NOPARMS m_DelegateOnGainFocus;
		DELEGATE_VOID_NOPARMS m_DelegateOnLoseFocus;

		Widget m_TraversalWidget;
		int m_TraversalDir;
		int m_MaxLength;

		bool m_InternalModify;
		bool m_ValidInvert;
		MISTRING m_ValidChars;
		MISTRING m_EmptyDisplay;

		static void CB_Activate (Widget, CTRL_EDIT_BASE*, void*);
		static void CB_Focus (Widget, CTRL_EDIT_BASE*, void*);
		static void CB_LosingFocus (Widget, CTRL_EDIT_BASE*, void*);
		static void CB_ModifyVerify (Widget, CTRL_EDIT_BASE*, void*);
		static void CB_ValueChanged (Widget, CTRL_EDIT_BASE*, void*);
		static void CB_Selection (Widget, CTRL_EDIT_BASE*, void*);

		void OnActivate ();
		void OnChangeSelection ();
		void OnUserEdit ();

		CTRL_EDIT_BASE (const CTRL_EDIT_BASE&);
		CTRL_EDIT_BASE& operator= (const CTRL_EDIT_BASE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		// CTRL_EDIT_BASE Overridables

		//! Called when selection is changed.
		virtual void v_OnChangeSelection ();

		//! Called when value is changed after validation.
		//! If the user modifies the text in the edit control this will not be called until the control loses "focus".
		virtual void v_OnChangeValue ();

		//! Called when user modifies the text in the edit control.
		//! The actual value does not get updated until the control loses focus or GetValue() is called.
		virtual void v_OnUserEdit ();

		//! Called internally to update control-specific data from edit control.
		virtual void v_UpdateValue (
			MISTRING& newvalue
			);

	};
#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(CTRL_EDIT_BASE::CREATEFLAGS)
#endif

//===================================================================================================================

}	// End namespace MGUI

#endif	// INC_MGUI_EDITBASE_H
