/**
 * \file mgui/toolmultipoint.h
 * \brief MGUI::TOOL_MULTIPOINT class definition
 *
 * \if NODOC
 * $Id: toolmultipoint.h_v 1.13 2011/12/09 20:23:12 mju Exp $
 *
 * $Log: toolmultipoint.h_v $
 * Revision 1.13  2011/12/09 20:23:12  mju
 * No context in v_draw.
 *
 * Revision 1.12  2011/06/21 14:13:22  mju
 * *** empty log message ***
 *
 * Revision 1.11  2008/05/02 20:02:18  scowan
 * Added the gps form.
 *
 * Revision 1.10  2008/04/28 23:04:56  scowan
 * Inherit from tool gps.
 *
 * Revision 1.9  2008/02/20 22:33:28  scowan
 * Finished manual entry panel.
 *
 * Revision 1.8  2008/02/19 00:21:20  scowan
 * More stable and somewhat usable manual entry panel in place.
 *
 * Revision 1.7  2008/02/01 15:01:13  mju
 * Change button handlers to return bool.
 *
 * Revision 1.6  2008/01/04 18:59:15  mju
 * Generalize pointer button methods.
 *
 * Revision 1.5  2007/12/21 22:06:46  mju
 * Add region generation.
 *
 * Revision 1.1  2003/01/10 23:08:25  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MGUI_TOOLMULTIPOINT_H
#define	INC_MGUI_TOOLMULTIPOINT_H

#ifndef	INC_MGUI_TOOLGPS_H
#include <mgui/toolgps.h>
#endif

#ifndef	INC_MI32_POLYLINE_H
#include <mi32/polyline.h>
#endif

namespace MGUI {

//=====================================================================================================================

//! Multi-Point tool.
class TOOL_MULTIPOINT : public TOOL_GPS {
	public:

		typedef fastdelegate::FastDelegate<bool(CHANGESEQ, DPOINT3D&)> DELEGATE_ONSETPOINT;
		
		//! Constructor.
		explicit TOOL_MULTIPOINT (
			TOOLCANVAS *canvas
			);

		//! Destructor.
		virtual ~TOOL_MULTIPOINT ();
		
		//! Add a point to the set
		void AddPoint (
			const DPOINT3D& pt,
			bool MakePointActive = true			//!< If not already, make this point the active point
			) { SetPoint(m_Points.GetNumPoints(), pt, MakePointActive); }
			
		//! Delete a single point.
		//! If the index is out of range the operation is ignored
		void DeletePoint (
			int index									//!< Zero-based index of point to delete
			);

		//! Get active point coordinates.
		DPOINT3D GetActivePoint (
			) const { return (GetPoint(m_ActiveIndex)); }

		//! Get active point array index.
		int GetActivePointIndex (
			) const { return (m_ActiveIndex); }

		//! Get number of points.
		int GetNumPoints (
			) const { return (m_Points.GetNumPoints()); }

		//! Get single point.
		//! If the index is out of range the results are unpredictable.
		DPOINT3D GetPoint (
			int index									//!< Zero-based index of point to retrieve
			) const;

		//! Get points.
		const POLYLINE& GetPoints (
			) const { return (m_Points); }

		//! Determine if "active" point is valid.
		bool IsActiveValid (
			) const { return (m_ActiveIndex >= 0 && m_ActiveIndex < m_Points.GetNumPoints()); }

		//! Set active point array index.
		void SetActivePointIndex (
			int index
			);

		//! Set delegate to call when a point is added or moved
		//! This allows the user of the tool to determine if the point is to be moved and where.
		//! Return 'true' for the delegate to modify the point.
		void SetDelegateOnSetPoint (
			DELEGATE_ONSETPOINT delegate
			) { m_DelegateOnSetPoint = delegate; }

		//! Set marker type and size.
		//! The default marker is MARKER_Plus with a size of 9 pixels.
		void SetMarker (
			MARKER marker,								//!< New marker type
			UINT16 size = 0							//!< New marker size, 0 to retain current size
			);

		//! Set single point.
		//! If the index is the same as the number of points, the point will be added
		//! otherwise if the index is out of range the point is ignored.
		void SetPoint (
			int index,									//!< Zero-based index of point to set
			const DPOINT3D& pt,						//!< Point to set
			bool MakePointActive = true			//!< If not already, make this point the active point
			);
			
		//! Set the list of points
		void SetPoints (
			const POLYLINE& PointArray
			);
			
	private:

		#ifndef GENERATING_DOXYGEN_OUTPUT

		enum EDITMODE {
			EDITMODE_Automatic,						//!< Auto select when near point, move if selected, add if not near point
			EDITMODE_Add,								//!< Always add new point
			EDITMODE_Move,								//!< Move selected point to cursor
			EDITMODE_Select							//!< Select point by clicking
			};

		enum ZONE {
			ZONE_NoFunction = 0,
			ZONE_Outside,
			};

		//! Tool settings
		MARKER m_marker;
		UINT16 m_markersize;
		EDITMODE m_EditMode;
		//! Tool position values
		POLYLINE m_Points;		//!< Points
		int m_ActiveIndex;						//!< Index of active point
		//! Values used during interaction
		ZONE m_CursorZone;						//!< Current cursor zone
		DPOINT2D m_prevpt;
		bool m_NotifyChangeElem;
		//! Delegates
		DELEGATE_ONSETPOINT m_DelegateOnSetPoint;

		//! Manual Point Entry form forward declaration
		class MANUAL_MULTIPOINT;
		friend class MANUAL_MULTIPOINT;		//!< Used to call the DELEGATE_ONSETPOINT delegate
		class FORM_MULTIPOINT;
		FORM_MULTIPOINT* m_pFormMultiPoint;

		//! Methods
		void ActivateClosestPoint (const DPOINT2D&);
		void AddPoint (MGD::CONTEXT*, const DPOINT2D&);
		void AdjustTool (DPOINT2D&, CHANGESEQ);
		void ComputeCursorZone (const DPOINT2D&);
		void DeletePoint ();
		void DrawNeighborPoints (MGD::CONTEXT*, const LRECT2D&, int SkipPoint) const;
		void DrawPoint (MGD::CONTEXT*, int) const;
		void ErasePoint (MGD::CONTEXT*, bool circleonly = false) const;
		void SetActiveIndex (int);
		void UpdateCursor () const;

		//! TOOL_BASE overrides.
		virtual MGUI::FORM* v_CreateControlForm (LAYOUT_PANE_BASE& ParentPane, CTRLOPT ctrlopt);
		virtual MANUAL_BASE* v_CreateManualBase ();
		virtual void v_DoDraw ();
		virtual void v_DoMove (double dx, double dy, ADJUSTFLAGS flags);
		virtual void v_OnClear ();
		virtual int v_GenerateRegion (REGION2D& region) const;
		virtual bool v_OnKey (MGUI::POINT point, KEYCODE keycode, KEYSTATE keystate);
		virtual bool v_OnPointerButtonDown (MGUI::POINT point, POINTERBUTTON button, KEYSTATE keystate);
		virtual bool v_OnPointerButtonUp (MGUI::POINT point, POINTERBUTTON button, KEYSTATE keystate);
		virtual void v_OnPointerMove (MGUI::POINT point, POINTERBUTTON button, KEYSTATE keystate);

		//! TOOL_GPS overrides.
		virtual void v_OnGPSCoordinate (const DPOINT3D& point);
		
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

}	// End namespace MGUI

#endif	// INC_MGUI_TOOLMULTIPOINT_H
