/*************************************************************************
 * \file mi32/dither.h
 * \brief Dither pattern classes.
 *
 * \if NODOC
 * $Log: dither.h_v $
 * Revision 1.13  2011/10/25 17:48:52  mju
 * Remove redundant hdr.
 *
 * Revision 1.12  2006/12/21 23:06:14  dwilliss
 * Decided to change the name
 *
 * Revision 1.11  2006/12/21 22:58:37  dwilliss
 * Added GetPatternLongName
 *
 * Revision 1.10  2006/04/20 15:41:02  dwilliss
 * made some virtual methods non-inline to silence unused parameter warnings
 *
 * Revision 1.7  2003/07/03 21:18:44  dwilliss
 * Added a dither line CMYK function
 *
 * Revision 1.4  2003/04/15 17:57:33  dwilliss
 * Added methods for setting dot gain, screen angles, and screen freq, as well
 * as methods to determine if a pattern supports any of those settings
 *
 * Revision 1.1  2002/10/22 20:56:56  dwilliss
 * Initial revision
 * \endif
 *****************************************************************************************/

#ifndef INC_MI32_DITHER_H
#define INC_MI32_DITHER_H

#ifndef INC_MI32_COLOR_H
#include <mi32/color.h>
#endif

#ifndef INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

class MISTRING;

class DITHER {
	public:

		//! Allocate a dither pattern
		static DITHER* Allocate (
			const char* PatternName,
			int numcolors,
			int cols,
			bool bForceMono
			);

		//! Get the name of a dither pattern given its pattern number
		//! Note, the name is in resource format (all lowercase, underscores, etc)
		static const char *GetPatternName (
			int num
			); 

		//! Get a dither pattern number given its name
		static int GetPatternNum (
			const char *name
			); 

		//! Bet the list of dither patterns
		static char **GetPatternList (
			int *countp, 
			bool bAllowErrorDiffusion = true 
			);

		static MISTRING GetPatternLongName (
			const char *name
			);

		virtual ~DITHER();

		//! Dither a solid color.  
		//!  The output is assumed to be 3 seperate buffers:C,M,Y with values
		//!  of 0 and 255
		virtual void DitherSolidColor (
			const COLOR& color,
			UINT8** out,
			int lin,
			int col, 
			int numcols
			);
		
		virtual bool CanUseCMYKcm (
			) const {
			return false;
			};

		//! Dither a solid color.  
		//! Input values are in the range of 0-255
		//!  The output is assumed to be 3 seperate buffers:C,M,Y with values
		//!  of 0 and 255
		//!  The version that takes a COLOR is prefered.  This version might
		//!  be deprecated some day
		virtual void DitherSolidColor (
			const UINT8 *rgb, 
			UINT8 **out, 
			int lin, 
			int col, 
			int cols, 
			UINT32 flags
			);

		//! Dither a solid gray
		void DitherSolidGray (
			UINT8 in,
			UINT8* out,
			int lin,
			int col,
			int numcols
			) {
			DitherSolidComponent(0, in, out, lin, col, numcols);
			}

		//!  Dither a line of CMYK data.
		//!  Input values are in the range of 0-255
		//!  in is expected to hold data like CMYKCMYK...
		//!  The output is assumed to be 3 seperate buffers:C,M,Y with values
		//!  of 0 and 255
		//!  Trans is a lookup table with 256 values which scales input values
		//!  into the range the pattern wants (0 to GetMaxVal()), with a possible
		//!  gama curve thrown in.  NOTE: is in RGB space, not CMYK space
		virtual void DitherLineCMYK (
			const UINT8* const* trans,
			const UINT8* in,
			UINT8** out,
			const UINT8* mask,
			int lin,
			int col, 
			int NumCols
			);

		//!  Dither a line of RGB data.
		//!  Input values are in the range of 0-255
		//!  in is expected to hold data like RGBRGBRGB...
		//!  The output is assumed to be 3 seperate buffers:C,M,Y with values
		//!  of 0 and 255
		//!  Trans is a lookup table with 256 values which scales input values
		//!  into the range the pattern wants (0 to GetMaxVal()), with a possible
		//!  gama curve thrown in.
		virtual void DitherLineRGB24 (
			const UINT8* const* trans,
			const UINT8* in,
			UINT8** out,
			const UINT8* mask,
			int lin,
			int col, 
			int NumCols
			);

		//! Dither a gray image line to a binary output raster.
		//!  Trans is a lookup table with 256 values which scales input values
		//!  into the range the pattern wants (0 to GetMaxVal()), with a possible
		//!  gama curve thrown in.
		virtual void DitherLineGray (
			const UINT8* trans,
			const UINT8* in,
			UINT8* out,
			const UINT8* mask,
			int lin,
			int col,
			int numcols
			) {
			DitherComponent(0, trans, in, out, mask, lin, col, numcols, 1);
			}
		


		//! Get the number of levels supported by the pattern
		int GetNumLevels (
			) const {
			return m_NumLevels;
			}

		//! Get the maximum value supported by the pattern
		int GetMaxVal (
			) const {
			return m_NumLevels - 1;
			}

		//! Return true if SetDotGain() is implemented
		//! If you override this, you must also implement SetDotGain()
		virtual bool HasDotGainCompensation (
			) const {
			return (false);
			}

		//! Return true if SetScreenAngle() is implemented
		//! If you override this, you must also implement SetScreenAngle()
		virtual bool HasVariableScreenAngle (
			) const {
			return (false);
			}

		//! Return true if SetScreenFreq() is implemented
		//! If you override this, you must also implement SetScreenFreq()
		virtual bool HasVariableScreenFreq (
			) const {
			return (false);
			}

		//! Set the ammount of "dot gain" for each componant
		//! If you override this, you should also override 
		//! 	HasDotGainCompensation() and make it return true
		virtual void SetDotGain (
			int comp,
			double size 		//!< 1.0 == "ideal", 1.2 == 20% than ideal, etc.
			);

		//! Set the screen angles
		//! If you override this, you should also override 
		//! 	HasVariableScreenAngle () and make it return true
		//! You should also call SetScreenAngle() in your Init() method to setup
		//! a default.
		virtual void SetScreenAngle (
			int comp,
			double angle		//!< screen angle in degrees
			);

		//! Set the screen interval
		//! If you override this, you should also override 
		//! 	HasVariableScreenFreq () and make it return true
		//! You should also call SetScreenFreq() in your Init() method to setup
		//! a default.
		virtual void SetScreenFreq (
			double HalftoneDPI,
			double PrinterDPI
			);

		//! Returns the ammount of area covered (0.0 to 1.0) at a given halftone
		//! level and DotGain.
		//! If DotGain is 1.0, this is trival.  If greater than 1.0, then it takes
		//! into account the size of the dots and overlap between adjacent dots.
		double ComputeCoverage (
			int level,			//!< range 0 to GetNumLevels()-1
			double DotGain		//!< 1.0="ideal",  1.1 means dot is 10% bigger than "ideal"
			);

		bool SetUseCMYKcm (
			bool val = true
			) {
			if (!CanUseCMYKcm()) return false;
			m_bUseCMYKcm = val;	
			return (val);
			};

	protected:

		//! Constructor is protected.  Can only call static method to allocate one.
		DITHER();

		#ifndef GENERATING_DOXYGEN_OUTPUT
		virtual void DitherComponent (
			int color,
			const UINT8* trans,
			const UINT8* in,
			UINT8* out,
			const UINT8* mask,
			int lin,
			int col,
			int numcols,
			int step
			) = 0;
		
		virtual void DitherSolidComponent (
			int color,
			UINT8 in,
			UINT8* out,
			int lin,
			int col,
			int numcols
			) = 0;

		virtual int Init (
			int PatternNum,
			int numcolors,
			int cols,
			bool bForceMono
			) = 0;

		void SetNumLevels (
			int n
			) {
			m_NumLevels = n;
			}

		bool m_bUseCMYKcm;
		#endif //!< GENERATING_DOXYGEN_OUTPUT

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		static int ReadPatternNames();

		int m_NumLevels;
		SIMPLE_ARRAY<UINT8> m_TempBuf;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};



#endif  // INC_MI32_DITHER_H
