/**
 * \file mi32/lidarlasfile.h
 * \brief LIDAR::LAS::FILE definitions.
 *
 * \if NODOC
 * $Id: lidarlasfile.h_v 1.13 2012/09/18 13:45:17 mju Exp $
 *
 * $Log: lidarlasfile.h_v $
 * Revision 1.13  2012/09/18 13:45:17  mju
 * Attempt dox fix.
 *
 * Revision 1.12  2012/09/13 21:23:10  mju
 * Add recordid_mi_BoundingRegion.
 *
 * Revision 1.11  2012/09/12 20:22:04  mju
 * Add methods to enable/disable multithread support.
 *
 * Revision 1.10  2012/09/07 20:25:08  mju
 * Add writeVariableRecord with void*.
 *
 * Revision 1.9  2012/08/31 21:15:03  mju
 * Simplify VLR handling by letting file keep track of what kind of vlrs are being used.
 * Support external vlrs in lasx file.
 *
 * Revision 1.8  2012/08/30 17:47:04  mju
 * Move userid and recordid enums outside file class.
 *
 * Revision 1.7  2012/06/19 17:19:52  mju
 * Add getPointRecordNumColorValues.
 *
 * Revision 1.6  2012/06/18 17:06:53  mju
 * Make getNumPointsByReturn non-inline.
 * Change setNumPointsByReturn to take number/count.
 *
 * Revision 1.5  2012/06/14 21:32:24  mju
 * Adjustments for LAS 1.4
 *
 * Revision 1.4  2011/07/15 17:29:38  mju
 * Add file::setupPointArray.
 *
 * Revision 1.1  2011/06/24 21:10:19  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_LIDARLASFILE_H
#define	INC_MI32_LIDARLASFILE_H

#ifndef	INC_MI32_FILEACCESS_H
#include <mi32/fileaccess.h>
#endif

#ifndef	INC_MI32_RECT_H
#include <mi32/rect.h>
#endif

#ifndef	INC_MI32_MISTRING
#include <mi32/mistring.h>
#endif

#ifndef	INC_MI32_RECORDBUFFER_H
#include <mi32/recordbuffer.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
struct COLOR;
class FILE_RAW;
class FILEPATH;
class QUANTIZER_LPOINT3D;
namespace SPATREF {
	class COORDREFSYS;
	}
namespace LIDAR {
	namespace LAS {
		class POINTARRAY;
		class POINTRECORD;
		class FILEHDR;
		}
	}
#endif


#ifdef RVCSYSDLL
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
#endif


namespace LIDAR {
namespace LAS {

//=====================================================================================================================

//! User ID enumeration for use with variable records.
enum USERID {
	USERID_Unknown =				0,
	USERID_LASF_Spec =			1,
	USERID_LASF_Projection =	2,
	USERID_MicroImages =			1172,		//!< Official ID number as provided by ASPRS
	USERID_MI_PUBLIC =			1281		//!< "Public" record formats
	};

//! Record ID enumeration for use with variable records.
enum RECORDID {
	RECORDID_LASF_Projection_WKT_Trans =				2111,
	RECORDID_LASF_Projection_WKT_CRS =					2112,
	RECORDID_LASF_Projection_GeoKeyDirectoryTag =	34735,
	RECORDID_LASF_Projection_GeoDoubleParamsTag =	34736,
	RECORDID_LASF_Projection_GeoAsciiParamsTag =		34737,
	RECORDID_LASF_Spec_ClassificationLookup =			0,
	RECORDID_LASF_Spec_TextAreaDescripton =			3,
	RECORDID_LASF_Spec_ExtraBytes =						4,
	RECORDID_LASF_Spec_Superseded =						7,
	RECORDID_LASF_Spec_WaveformPacketDescBase =		100,
	RECORDID_LASF_Spec_WaveformPacketDescMax =		354,
	RECORDID_LASF_Spec_WaveformData =					65535,
	// These can occur in either "MicroImages" or "MI_PUBLIC" VLRs
	RECORDID_MI_TimeStamp =									0,
	RECORDID_MI_ClassHistogram =							1,
	RECORDID_MI_FileStats =									2,
	RECORDID_MI_BadPointSet =								3,
	RECORDID_MI_BoundingRegion =							4,
	RECORDID_MI_Grid_RangeZ =								100,
	RECORDID_MI_Grid_Count =								101,
	RECORDID_MI_Grid_ReturnType =							102,
	RECORDID_MI_Grid_Index =								103,
	RECORDID_MI_Histogram_Intensity =					200,
	RECORDID_MI_Histogram_RawZ =							201,
	};

enum DATATYPE {
	DATATYPE_Undefined =			0,
	DATATYPE_UINT8 =				1,
	DATATYPE_SINT8 =				2,
	DATATYPE_UINT16 =				3,
	DATATYPE_SINT16 =				4,
	DATATYPE_UINT32 =				5,
	DATATYPE_SINT32 =				6,
	DATATYPE_UINT64 =				7,
	DATATYPE_SINT64 =				8,
	DATATYPE_FLOAT32 =			9,
	DATATYPE_FLOAT64 =			10,
	};

//=====================================================================================================================

//! Container for VLR location.
class VLRLOC {
	public:

		//! Place within file for VLR.
		enum PLACE {
			PLACE_SVLR = 0,	//!< In "short VLR" area of LAS file
			PLACE_EVLR,			//!< In "extended VLR" area of LAS file
			PLACE_LASX,			//!< In LASX file
			PLACE_None			//!< No location, record does not exist
			};

		//! Constructor.
		VLRLOC (
			) : m_place(PLACE_None), m_index(UINT32_MAX) { }

		//! Constructor.
		VLRLOC (
			PLACE place,
			UINT32 index
			) : m_place(place), m_index(index) { }

		//! Get index.
		UINT32 GetIndex () const
			{ return (m_index); }

		//! Get place.
		PLACE GetPlace () const
			{ return (m_place); }

		//! Set location.
		void Set (
			PLACE place,
			UINT32 index
			) { m_place = place; m_index = index; }

		//! Set index.
		void SetIndex (
			UINT32 index
			) { m_index = index; }

		//! Set place.
		//! Index will be set to UINT32_MAX to indicate invalid.
		void SetPlace (
			PLACE place
			) { m_place = place; m_index = UINT32_MAX; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		PLACE m_place;
		UINT32 m_index;
		#endif
	};

//=====================================================================================================================

//! Variable-length record description.
class CLASSLIBEXPORT VLRDESC {
	public:

		//! Convert user ID string to enum.
		static USERID UserIdStrToEnum (
			const char *UserIdStr					//!< User ID string
			);

		//! Get user ID string from enum.
		//! @return string to store in LAS file.
		static const char* UserIdEnumToStr (
			USERID userid								//!< Use ID enum
			);

		//! Constructor.
		VLRDESC ();

		//! Create with specified values.
		ERRVALUE Create (
			const char *UserIdStr,
			UINT16 RecordID,
			UINT64 RecordLen,
			const char *Desc = 0
			);

		//! Create with specified values.
		ERRVALUE Create (
			USERID UserID,
			UINT16 RecordID,
			UINT64 RecordLen,
			const char *Desc = 0
			);

		//! Get description string.
		const char* GetDescription () const
			{ return (m_Description); }

		//! Get 'Record ID'.
		UINT16 GetRecordID () const
			{ return (m_RecordID); }

		//! Get record length in bytes, does not include header.
		UINT64 GetRecordLength () const
			{ return (m_RecordLength); }

		//! Get UserID enumeration.
		USERID GetUserID () const
			{ return (UserIdStrToEnum(m_UserIdStr)); }

		//! Get 'User ID' string, may not be 0-terminated if uses all 16 possible characters.
		const char* GetUserIdStr () const
			{ return (m_UserIdStr); }

		//! Determine if has same IDs as another VLRDESC.
		bool HasSameIds (
			const VLRDESC& rhs
			) const;

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		UINT64 m_RecordLength;
		UINT16 m_RecordID;
		char m_UserIdStr[16];
		char m_Description[32];
		#endif
	};

//=====================================================================================================================

//! File header.
class CLASSLIBEXPORT FILEHDR {
	public:

		//! GPS Time Type in point records.
		enum POINTGPSTIMETYPE {
			POINTGPSTIMETYPE_None,			//!< No GPS Time field
			POINTGPSTIMETYPE_Week,			//!< GPS Week Time
			POINTGPSTIMETYPE_StdAdj			//!< Standard GPS Time minus 1E+9 seconds
			};

		FILEHDR ();

		//! Add point record information to header.
		//! Does not update file itself.
		void AddPointRecord (
			const POINTRECORD& pointrec
			);

		//! Add variable record information to header.
		//! Does not update file itself.
		void AddVariableRecord (
			const VLRLOC& vlrloc,
			const VLRDESC& vlrdesc,
			UINT64 fileoffset
			);

		//! Compute raw extents.
		void ComputeExtentsRaw (
			LRECT3D& ExtentsRaw			//!< Raw extents returned
			) const;

		//! Get file creation day of year.
		UINT16 GetCreationDayOfYear () const
			{ return (m_CreationDayOfYear); }

		//! Get file creation year.
		UINT16 GetCreationYear () const
			{ return (m_CreationYear); }

		MISTRING GetDisplayValues () const;

		UINT32 GetFileOffsetPointData () const
			{ return (m_FileOffsetPointData); }

		//! Get extents in DRECT2D.
		void GetExtents (
			DRECT2D& extents
			) const { extents.Set(m_MinX,m_MinY,m_MaxX,m_MaxY); }

		//! Get extents in DRECT3D.
		void GetExtents (
			DRECT3D& extents
			) const { extents.Set(m_MinX,m_MinY,m_MinZ,m_MaxX,m_MaxY,m_MaxZ); }

		//! Get file source ID.
		UINT16 GetFileSourceID () const
			{ return (m_FileSrcID); }

		UINT16 GetHeaderSize () const
			{ return (m_HeaderSize); }

		//! Get maximum X value.
		double GetMaxX () const
			{ return (m_MaxX); }

		//! Get maximum Y value.
		double GetMaxY () const
			{ return (m_MaxY); }

		//! Get maximum Z value.
		double GetMaxZ () const
			{ return (m_MaxZ); }

		//! Get minimum X value.
		double GetMinX () const
			{ return (m_MinX); }

		//! Get minimum Y value.
		double GetMinY () const
			{ return (m_MinY); }

		//! Get minimum Z value.
		double GetMinZ () const
			{ return (m_MinZ); }

		//! Get number of extended variable-length records.
		UINT32 GetNumExtendedVarRecords () const
			{ return (m_NumExtendedVLRs); }

		//! Get total number of points.
		UINT64 GetNumPoints () const
			{ return (m_NumPoints); }

		//! Get number of points by return.
		UINT64 GetNumPointsByReturn (
			UINT8 retnum						//!< Return number (1-15)
			) const;

		//! Get number of short variable-length records.
		UINT32 GetNumShortVarRecords () const
			{ return (m_NumShortVLRs); }

		UINT64 GetOffsetExtendedVLRs () const
			{ return (m_OffsetExtendedVLRs); }

		//! Get point record format ID.
		UINT8 GetPointFormat () const
			{ return (m_PointDataFormatID); }

		//! Get point offset.
		const DPOINT3D& GetPointOffset () const
			{ return (m_Offset); }

		const WPOINT3D& GetPointPrecision () const
			{ return (m_PointPrecision); }

		UINT16 GetPointRecordNumColorValues () const;

		UINT16 GetPointRecordNumFields () const
			{ return (m_PointRecNumFields); }

		//! Get point record size in bytes.
		UINT16 GetPointRecordSize () const
			{ return (m_PointRecordSize); }

		//! Get point scale.
		const DPOINT3D& GetPointScale () const
			{ return (m_Scale); }

		//! Get project ID as string.
		MISTRING GetProjectID () const;

		//! Get point quantizer.
		void GetQuantizer (
			QUANTIZER_LPOINT3D& quantizer
			) const;

		//! Get "Generating Software" string.
		MISTRING GetSoftware () const;

		//! Get "System Identifier" string.
		MISTRING GetSystemID () const;

		//! Get major version number.
		UINT8 GetVersionMajor () const
			{ return (m_VersionMajor); }

		//! Get minor version number.
		UINT8 GetVersionMinor () const
			{ return (m_VersionMinor); }

		bool HasPointRecordColor () const;
		bool HasPointRecordTime () const;
		bool HasPointRecordWaveform () const;

		bool IsChanged () const
			{ return (m_IsChanged); }

		ERRVALUE Read (FILE_RAW& file);

		//! Set extents.
		void SetExtents (
			const DRECT3D& extents
			);

		//! Set number of points by return.
		void SetNumPointsByReturn (
			UINT8 retnum,						//!< Return number (1-15)
			const UINT64 NumPoints			//!< Number of points
			);

		void SetPointFormat (
			UINT8 PointFormat
			);

		void SetPointScaleOffset (
			const DPOINT2D& Scale,
			const DPOINT2D& Offset
			) { m_Scale.x = Scale.x; m_Scale.y = Scale.y; m_Offset.x = Offset.x; m_Offset.y = Offset.y; }

		void SetPointScaleOffset (
			const DPOINT3D& Scale,
			const DPOINT3D& Offset
			) { m_Scale = Scale; m_Offset = Offset; }

		void SetPointScaleOffsetZ (
			double ScaleZ,
			double OffsetZ
			) { m_Scale.z = ScaleZ; m_Offset.z = OffsetZ; }

		void SetSoftware (
			const char *Software
			);

		void SetSystemID (
			const char *SystemID
			);

		ERRVALUE SetupNew (const FILEHDR& srchdr);

		void SetVersion (
			UINT8 VersionMajor,
			UINT8 VersionMinor
			);

		ERRVALUE UpdateExtents (
			const FILEHDR& srchdr,
			FILE_RAW& fileraw
			);

		ERRVALUE UpdateNumPointsByReturn (
			const FILEHDR& srchdr,
			FILE_RAW& fileraw
			);

		void ValidateOffsetInExtents (
			POINTARRAY& PointRecArray
			);

		ERRVALUE Write (FILE_RAW& file);

		ERRVALUE WriteChanges (FILE_RAW& file);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		// Values written to file, order is important.
		char m_Signature[4];
		UINT16 m_FileSrcID;
		UINT16 m_GlobalEncoding;
		UINT32 m_GUID1;
		UINT16 m_GUID2;
		UINT16 m_GUID3;
		UINT8 m_GUID4[8];
		UINT8 m_VersionMajor;
		UINT8 m_VersionMinor;
		char m_SystemID[32];
		char m_OrigSoftware[32];
		UINT16 m_CreationDayOfYear;
		UINT16 m_CreationYear;
		UINT16 m_HeaderSize;
		UINT32 m_FileOffsetPointData;
		UINT32 m_NumShortVLRs;
		UINT8 m_PointDataFormatID;
		// 105 bytes to here, above read as unit, below read separately
		UINT16 m_PointRecordSize;
		// 12 doubles below read as unit
		DPOINT3D m_Scale;
		DPOINT3D m_Offset;
		double m_MaxX;
		double m_MinX;
		double m_MaxY;
		double m_MinY;
		double m_MaxZ;
		double m_MinZ;
		// Added at version 1.3
		UINT64 m_OffsetWaveData;
		// Added at version 1.4
		UINT64 m_OffsetExtendedVLRs;
		UINT32 m_NumExtendedVLRs;
		UINT64 m_NumPoints;
		UINT64 m_NumPointsByReturn[15];

		// Other calculated values not written to file
		WPOINT3D m_PointPrecision;
		UINT16 m_PointRecNumFields;
		POINTGPSTIMETYPE m_PointTimeType;
		bool m_IsChanged;

		void CalcNumFields ();
		void CalcPrecision ();
		void Clear ();
		void SwapBytes ();
		#endif			
	
	};

//=====================================================================================================================
//! Support for reading and writing LAS format.
//! This class manages shared access to the same physical file within a single process so that all instances
//! that reference that file will have correct data after any updates to the file.
class CLASSLIBEXPORT FILE {
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		class PRIV;
		#endif
	public:

		//! Iterator for Variable Length Records.
		class CLASSLIBEXPORT ITERATOR_VLR {
			public:

				//! Construct for specified LAS::FILE.
				explicit ITERATOR_VLR (
					FILE& file
					);

				#ifndef GENERATING_DOXYGEN_OUTPUT
				explicit ITERATOR_VLR (
					FILE::PRIV *pPriv
					);
				#endif

				//! Destructor.
				~ITERATOR_VLR ();

				//! Dereference as pointer to VLRDESC.
				const VLRDESC* operator-> () const
					{ return (&m_vlrdesc); }

				//! Dereference as reference to VLRLOC.
				const VLRLOC& operator* () const
					{ return (m_vlrloc); }

				//! Pre-increment operator.
				ITERATOR_VLR& operator++ ();

				const VLRDESC& GetDesc () const
					{ return (m_vlrdesc); }

				//! Determine if done with all VLRs.
				bool IsDone () const
					{ return (m_vlrloc.GetPlace() == VLRLOC::PLACE_None); }

				//! Reset to beginning.
				void Reset ();

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				FILE::PRIV& m_priv;
				VLRLOC m_vlrloc;
				VLRDESC m_vlrdesc;
				#endif
			};

		//! Disable multithread support for FILE access.
		static void DisableMultiThreading ();

		//! Enable multithread support for FILE access.
		//! If used, must be done before accessing FILE from any background thread.
		static void EnableMultiThreading ();

		//! Determine if file is supported format.
		//! Used as delegate for FILETYPE.
		static bool FileTypeIsFormat (FILE_RAW& file);

		//! Register file type in common set.
		static void FileTypeRegister ();

		//! Default constructor.
		FILE ();

		//! Destructor.
		~FILE ();

		//! Add point record.
		ERRVALUE AddPointRecord (
			const POINTRECORD& pointrec				//!< Point record to add
			);

		//! Add variable record.
		ERRVALUE AddVariableRecord (
			VLRLOC& vlrloc,								//!< Location passed/returned, if specifies PLACE_LASX will force to be in "extended" file
			const VLRDESC& vlrdesc,						//!< Record description
			const RECORDBUFFER_LSBFIRST& buffer		//!< Buffer to write
			);

		//! Close file.
		void Close ();

		//! Delete variable record.
		ERRVALUE DeleteVariableRecord (
			const VLRLOC& vlrloc							//!< Location of record to delete
			);

		//! Find first VLR with specified User/Record ID combination.
		//! @return TRUE if found, FALSE if not, < 0 if error.
		int FindVariableRecord (
			USERID UserID,									//!< User ID
			UINT16 RecordID,								//!< Record ID
			VLRLOC& vlrloc									//!< VLR location returned if found
			);

		//! Get file header.
		const FILEHDR& GetHeader () const;

		//! Get number of point records.
		UINT64 GetPointRecordCount () const
			{ return (GetHeader().GetNumPoints()); }

		//! Get format of point records.
		UINT8 GetPointFormat () const
			{ return (GetHeader().GetPointFormat()); }

		//! Get precision (number of decimal places) for point coordinates.
		const WPOINT3D& GetPointPrecision () const
			{ return (GetHeader().GetPointPrecision()); }

		//! Get number of color value fields in point record.
		UINT16 GetPointRecordNumColorValues () const
			{ return (GetHeader().GetPointRecordNumColorValues()); }

		//! Get number of fields in point record.
		UINT16 GetPointRecordNumFields () const
			{ return (GetHeader().GetPointRecordNumFields()); }

		//! Get size of point record in bytes.
		UINT16 GetPointRecordSize () const
			{ return (GetHeader().GetPointRecordSize()); }

		//! Determine if has "color" field in point record.
		bool HasPointRecordColor () const
			{ return (GetHeader().HasPointRecordColor()); }

		//! Determine if has "time" field in point record.
		bool HasPointRecordTime () const
			{ return (GetHeader().HasPointRecordTime()); }

		//! Determine if has "waveform" fields in point record.
		bool HasPointRecordWaveform () const
			{ return (GetHeader().HasPointRecordWaveform()); }

		//! Determine if file is open.
		bool IsOpen () const
			{ return (m_pPriv != 0); }

		//! Determine if file is writable.
		bool IsWritable () const
			{ return (m_IsWritable); }

		//! Make new LAS file.
		ERRVALUE Make (
			const FILEPATH& filepath,
			const FILEHDR& header					//!< Header template, extents and point count fields will be ignored
			);

		//! Open LAS file.
		ERRVALUE Open (
			const FILEPATH& filepath,
			FILE_ACCESS access = FILE_ACCESS_Read			//!< Desired access, FILE_ACCESS_Read or FILE_ACCESS_ReadWrite
			);

		//! Read Coordinate Reference System information if available.
		ERRVALUE ReadCoordRefSys (
			SPATREF::COORDREFSYS& CoordRefSys	//!< CRS returned, will be "undefined" if no info in file
			);

		//! Read point records.
		//! POINTRECORD instances that reference the specified POINTARRAY will be overwritten by new data or become invalid.
		ERRVALUE ReadPointRecords (
			POINTARRAY& pointrecs,					//!< Point records returned, will be attached/resized if necessary
			UINT32 NumPoints,							//!< Number of points to read
			UINT64 FirstPointIdx = UINT64_MAX	//!< Index of first point to read, UINT64_MAX to continue from previous read
			);

		//! Read entire variable record data into buffer.
		ERRVALUE ReadVariableRecordData (
			const VLRLOC& vlrloc,					//!< Record location
			RECORDBUFFER_LSBFIRST& buffer			//!< Buffer to contain result
			);

		//! Read variable record data with offset/size.
		ERRVALUE ReadVariableRecordData (
			const VLRLOC& vlrloc,					//!< Record location
			void *data,									//!< Container for result
			UINT64 offset,								//!< Offset within record
			UINT32 numbytes							//!< Number of bytes to read
			);

		//! Read variable record description.
		ERRVALUE ReadVariableRecordDesc (
			const VLRLOC& vlrloc,					//!< Record location
			VLRDESC& vlrdesc							//!< Container for result
			);

		//! Setup point array associated with file.
		//! Done automatically when reading.
		ERRVALUE SetupPointArray (
			POINTARRAY& PointArray,
			UINT32 NumPoints
			);

		//! Update extents to file from specified header.
		ERRVALUE UpdateHeaderExtents (
			const FILEHDR& header
			);

		//! Update "number of points by return" to file from specified header.
		ERRVALUE UpdateHeaderNumPointsByReturn (
			const FILEHDR& header
			);

		//! Write CRS variable records.
		ERRVALUE WriteCoordRefSys (
			const SPATREF::COORDREFSYS& crs
			);

		//! Write point record.
		ERRVALUE WritePointRecord (
			const POINTRECORD& pointrec,			//!< Point record to write
			UINT64 PointIdx							//!< Index of point to write
			);

		//! Write complete variable record.
		//! Record will be added if no record already exists with specified user/record IDs.
		ERRVALUE WriteVariableRecord (
			VLRLOC& vlrloc,								//!< Location passed/returned, if specifies PLACE_LASX will force to be in "extended" file
			const VLRDESC& vlrdesc,						//!< Record description
			const RECORDBUFFER_LSBFIRST& buffer		//!< Buffer to write
			);

		//! Write complete variable record.
		//! Record will be added if no record already exists with specified user/record IDs.
		ERRVALUE WriteVariableRecord (
			VLRLOC& vlrloc,								//!< Location passed/returned, if specifies PLACE_LASX will force to be in "extended" file
			const VLRDESC& vlrdesc,						//!< Record description
			const void *data								//!< Data to write, size specified by VLRDESC
			);

		//! Write into existing variable record data with offset/size.
		ERRVALUE WriteVariableRecordData (
			const VLRLOC& vlrloc,						//!< Location
			const void *data,								//!< Data to write
			UINT64 offset,									//!< Offset within record
			UINT32 numbytes								//!< Number of bytes to Write
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		PRIV *m_pPriv;
		UINT64 m_PointIdxNextRead;
		bool m_IsWritable;

		// Unimplemented.
		FILE (const FILE&);
		FILE& operator= (const FILE&);
		#endif
	};

//=====================================================================================================================

}	// End namespace LAS
}	// End namespace LIDAR

#undef	CLASSLIBEXPORT

#endif	// INC_MI32_LIDARLASFILE_H
