/**
 * \file mi32/migrid.h
 * \brief Definitions for MIGRIDCTRL
 *
 * \if NODOC
 * $Id: migrid.h_v 1.57 2011/12/02 15:39:12 mju Exp $
 *
 * $Log: migrid.h_v $
 * Revision 1.57  2011/12/02 15:39:12  mju
 * Remove legacy hdr.
 *
 * Revision 1.56  2011/11/10 14:47:13  mju
 * Use ucstring hdr.
 *
 * Revision 1.53  2007/06/25 17:50:40  dwilliss
 * Found a better solution for what I was trying to do. v_DrawText now has
 * optional parameter to override the text
 *
 * Revision 1.49  2006/11/08 18:37:59  dwilliss
 * right mouse button menu now implemented through MGUI::MENU.
 * v_OnContextMenuStart gets passed the MGUI::MENU (there's still a version
 * which doesn't so old code will still work.  Default implementation of hte
 * [the] new method calls the old one)
 *
 * Revision 1.47  2006/11/03 23:50:03  dwilliss
 * Changed to use MGUI tooltip
 *
 * Revision 1.42  2006/04/05 14:55:41  mju
 * Use ptrint/ptruint for get/setItemData.
 *
 * Revision 1.32  2002/06/07 15:40:52  dwilliss
 * Added GetColumnOrderArray
 *
 * Revision 1.30  2002/06/06 16:27:55  dwilliss
 * On resize, reposition drawing area if would otherwise leave part hidden and
 * only expose nothingness to the right
 *
 * Revision 1.28  2002/03/15 22:34:26  dwilliss
 * Changed the way the imagelist function works.  Setting an image list now
 * defaults to copying the list
 *
 * Revision 1.27  2002/02/19 20:21:14  dwilliss
 * OnEditStart can now actually cancel the edit
 *
 * Revision 1.17  2000/11/15 14:44:42  mju
 * Add standard header comment block.
 * Add inclusion guards.
 * \endif
**/

#ifndef	INC_MI32_MIGRID_H
#define	INC_MI32_MIGRID_H

#ifndef INC_MI32_MGD2_H
#include <mi32/mgd2.h>
#endif

#ifndef INC_MI32_MEMPOOL_H
#include <mi32/mempool.h>
#endif

#ifndef INC_MGUI_TOOLTIP_H
#include <mgui/tooltip.h>
#endif

#ifndef INC_MGUI_MENU_H
#include <mgui/menu.h>
#endif

#ifndef INC_MI32_UCSTRING_H
#include <mi32/ucstring.h>
#endif

#ifndef INC_MI32_XDEFNS_H
#include <mi32/xdefns.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
class MIGRIDCELL;
class MIGRIDCTRL;
class MIGRIDEDITHANDLE;
class DBEDIT;
class DBEDIT_TDATA;
namespace MGUI {
	class IMAGELIST;
	class CTRL_TOOLTIP;
	}
namespace MGD {
	class IMAGELIST;
	class DEVICE_X;
	}
#endif


//////////////////////////////////////////////////////////////////////////////!
//////////////////////////!                          //////////////////////////
///////////////////////!           MIGRIDCELL          ////////////////////////
//////////////////////////!                          //////////////////////////
//////////////////////////////////////////////////////////////////////////////!

class MIGRIDCELL {
public:
	enum ALIGNMENT {
		ALIGN_Left		 = 0,
		ALIGN_Right,
		ALIGN_Center
		};
	
	enum BORDER {
		BORDER_None			= 0x0000,		
		BORDER_LThin		= 0x0001,
		BORDER_TThin		= 0x0002,
		BORDER_RThin		= 0x0004,
		BORDER_BThin   	= 0x0008,
		BORDER_Thin    	= 0x000F,
		BORDER_LMedium		= 0x0010,
		BORDER_TMedium		= 0x0020,
		BORDER_RMedium		= 0x0040,
		BORDER_BMedium 	= 0x0080,
		BORDER_Medium   	= 0x00FF,
		BORDER_LThick		= 0x0100,
		BORDER_TThick		= 0x0200,
		BORDER_RThick		= 0x0400,
		BORDER_BThick  	= 0x0800,
		BORDER_Thick   	= 0x0FFF,
		BORDER_Left  		= 0x0111,		//!< Mask for left, any size
		BORDER_Top			= 0x0222,		//!< Mask for top, any size
		BORDER_Right  		= 0x0444,		//!< Mask for Right, any size
		BORDER_Bottom  	= 0x0888,		//!< Mask for bottom, any size
		BORDER_Recessed	= 0x1000,
		BORDER_Raised		= 0x2000
		};	


	enum PROPERTY {
		PROP_None			= 0x00000000,
		PROP_Text			= 0x00000001,
		PROP_FGColor		= 0x00000002,
		PROP_BGColor		= 0x00000004,
		PROP_BorderStyle	= 0x00000008,
		PROP_Alignment		= 0x00000010,
		PROP_DecimalPlaces= 0x00000020,
		PROP_ReadOnly		= 0x00000040,
		PROP_DataType		= 0x00000080,
		PROP_Param			= 0x00000100,
		PROP_CellType		= 0x00000200,
		PROP_CellTypeEx	= 0x00000400,
		PROP_Label			= 0x00000800,
		PROP_FGHColor		= 0x00001000,
		PROP_BGHColor		= 0x00002000,
		PROP_Bool			= 0x00004000,
		PROP_MaxLen			= 0x00008000,
		PROP_Icon			= 0x00010000,
		PROP_All				= 0xFFFFFFFF	//!< All properties
		};

	enum DATATYPE {
		DATATYPE_String,
		DATATYPE_Number,
		DATATYPE_Bool,
		DATATYPE_Time,
		DATATYPE_Currency,
		DATATYPE_Color
		};

	//! These are the built in celltypes.  Note that calling MIGRIDCTRL::AddCellType()
	//! will return a value not in this list.  That's O.K.!  This is an enum for
	//! type safety.
	//!
	//!	Special cases:
	//!		CELLTYPE_Text
	//!			OnEditFinish will be called when a cell is edited and loses focus.
	//!
	//!		CELLTYPE_Button
	//!			If cell.SetReadOnly(true), button looks normal but can't be pressed.
	//!			To make a button a toggle, call cell.SetCellTypeEx(CELLTYPEEX_Toggle)
	//!			When the button is pressed, OnCellTypeNotify will be called.  If 
	//!			the button is a toggle, the "param" parameter of OnCellTypeNotify()
	//!			will be 1 for on, 0 for off.  You can turn a toggle on/off by calling
	//!			SetBool() on the cell.
	//!			
	//!		CELLTYPE_DropDown
	//!		CELLTYPE_PopUp
	//!			Set the list of values for the dropdown by calling cell.SetLabelText()
	//!			(See details below).  OnEditStart() is called before the list is
	//!			created, so you can delay populating it until then if you wish.
	//!			[ Handy tip.  If all cells in a column use the same list, set it
	//!			with SetColDefault.  In OnEditStart, call GetColDefault(), and only
	//!			bother generating the list if GetLabelText() returns NULL ]
	//!
	//!		CELLTYPE_Icon
	//!			Call cell.SetLabelText("group.icon") to set the icon.  To make the
	//!			icon stippled, call cell.SetCellTypeEx(CELLTYPEEX_Disabled)
	//!			You can actually give the label two icons by passing two names
	//!			like cell.SetLabelText("group.icon1, group.icon2") and switch 
	//!			between them by calling cell.SetBool() (false is the first one,
	//!			true is the second)
	//!
	//!		CELLTYPE_IconButton
	//!			A combination of CELLTYPE_Button and CELLTYPE_Icon.  The 2nd icon
	//!			is used for the "down" state.
	//!
	//!		CELLTYPE_Color
	//!			A special type of CELLTYPE_Button.  The Cell's ForeColor is used 
	//!			for the color of the button.  Clicking on the button (if not set
	//!			to ReadOnly) pops up a color editor and calls EditFinish()
	//!			
	enum CELLTYPE {
		CELLTYPE_Text = 0,					//!< Simple text cell (default)
		CELLTYPE_Button = 1,					//!< Simple pushbutton
		CELLTYPE_DropDown = 2,				//!< Dropdown list
		CELLTYPE_Icon = 3,						
		CELLTYPE_IconButton = 4,
		CELLTYPE_BigText = 5,		 		//!< Text with [...] button to see/edit whole thing.
		CELLTYPE_PopUp = 6,					//!< Popup list
		CELLTYPE_Color = 7					//!< Button with color sample and no text. 
		};

	//! Some of the celltypes have extra parameters that can be set on the cells.
	//! Call SetCellTypeEx with any of the following...
	enum {
		CELLTYPEEX_Toggle		= 0x00000001,		//!< For Button and Icon, make it a toggle button (Use GetBool/SetBool to get/set state)
		CELLTYPEEX_Disabled	= 0x00000002		//!< Draw icon or text stippled
		};

	// CONSTRUCTION / DESTRUCTION
	
	//! Constructor.
	MIGRIDCELL (
		);

	//! Constructor.
	MIGRIDCELL (
		const MIGRIDCELL&
		);

	//! Destructor.
	~MIGRIDCELL(
		);

	// OPERATORS

	//! Assignment operator.
	MIGRIDCELL& operator=(
		const MIGRIDCELL&
		);

	// METHODS

	//! Clear all properties to the "unset" state.
	void ClearAll (
		);

	//! Clear given properties.
	void ClearProperty (
		PROPERTY p
		);

	//! Get the alignment.
	//!
	//! @return the alignment.
	ALIGNMENT GetAlignment (
		) const {
		return (m_set & PROP_Alignment) ? m_Alignment : ALIGN_Left;
		}

	//! Get background color.
	const COLOR& GetBackColor (
		) const {
		return (m_set & PROP_BGColor) ? m_BGColor : s_black;
		}
		
	//! Get background color while the cell is hilighted.
	const COLOR& GetBackColorHilighted (
		) const	{
		return (m_set & PROP_BGHColor) ? m_BGHColor : GetBackColor();
		}

	//! Gets a boolean value setable via SetBool().
	bool GetBool (
		) const 
		{
		return (m_set & PROP_Bool) ? m_bState : false;
		}
		
	//! Get the border.
	//!
	//! @return The border.
	BORDER GetBorder (
		) const {
		return (m_set & PROP_BorderStyle) ? m_BorderStyle : BORDER_Thin;
		}

	//! Get the "cell type".
	//!
	//! @return The "cell type".
	//!
	//! See SetCellType() for details.
	CELLTYPE GetCellType (
		) const {
		return (m_set & PROP_CellType) ? m_CellType : CELLTYPE_Text;
		}

	//! Get the extra data value for the current cell type.
	//!
	//! @return The extra data value for the current cell type.
	int GetCellTypeEx (
		) const {
		return (m_set & PROP_CellTypeEx) ? m_nCellTypeEx : 0;
		}

	//! Get the data type for the cell.
	//!
	//! @return The data type for the cell.
	DATATYPE GetDataType (
		) const {
		return (m_set & PROP_DataType) ? m_DataType : DATATYPE_String;
		}

	//! Retrieves the color used for drawing.
	//!
	//! @return The color used for drawing.
	const COLOR& GetForeColor (
		) const {
		return (m_set & PROP_FGColor) ? m_FGColor : s_black;
		}

	//! Retrieves the color used for drawing while the cell is hilighted.
	//!
	//! @return The color used for drawing while the cell is hilighted.
	//!
	//! UGGrid calls this GetHTextColor().
	const COLOR& GetForeColorHilighted (
		) const {
		return (m_set & PROP_FGHColor) ? m_FGHColor : GetBackColor();
		}

	//! Return the index of the icon associated with this entry.
	//! 
	//! Returns the index into the IMAGELIST for this grid.  The IMAGELIST
	//! Is set with the MIGRIDCTRL::SetImageList() method.
	//! If no icon is set, returns -1.
	//! If an item has an icon, it will be displayed to the left of its
	//! content.
	int GetIconIndex (
		) const {
		return (m_set & PROP_Icon) ? m_IconIndex : -1;
		}

	//! Retrieves a const pointer to the label text.
	//! 
	//! @return A const pointer to the label text. 
	//! This is NOT the contents of the field.  See SetLabelText() for details
	const MIUNICODE* GetLabelText (
		) const {
		return (m_set & PROP_Label) ? m_Label : NULL;
		}


	//! Get the maximum length of the field, (-1) if no max set.
	//!
	//! @return The maximum length or -1 if no max is set.
	int GetMaxLen (
		) const {
		return (m_set & PROP_MaxLen) ? m_nMaxLen : -1;
		}

	//! Retrieves the value of the cell as a number.
	//!
	//! @return The value of the cell as a number.
	//!
	//! If the cell is a string cell, it converts it to a number.  An empty cell will
	//! return IEEE_NaN.
	double GetNumber (
		) const;

	//! Get the number of decimal places to show on numeric fields.
	//!
	//! @return The number of decimal places to show on mumeric fields.
	int GetNumberPlaces (
		) const {
		return (m_set & PROP_DecimalPlaces) ? m_nPlaces : 0;
		}


	//! Gets a general-usage UserData pointer which can be set with SetParam.
	void* GetParam (
		) const {
		return (m_set & PROP_Param) ? m_pParam : NULL;
		}

	//! Sees if is read only.
	bool GetReadOnly (
		) const {
		return (m_set & PROP_ReadOnly) ? m_bReadOnly : false;
		}

	//! Get pointer to the text.
	//! 
	//! @return A const pointer to the text. 
	const MIUNICODE* GetText (
		) const {
		static MIUNICODE EmptyString = 0;
		return ((m_set & PROP_Text) && m_str) ? m_str : &EmptyString;
		}

	//! Retrieves a copy of the string which you must free.
	//!
	//! @return A copy of the string which you must free, but it's not const.
	MIUNICODE* GetTextCopy (
		) const {
		return ((m_set & PROP_Text) && m_str) ? ucstrdup(m_str) : strtoucdup("");
		}

	//! See if cell has a value.
	//!
	//! @return True if the cell has no value.
	bool IsEmpty (
		) const {
		if (!(m_set & PROP_Text)) return (false);
		return (m_str[0] == 0);
		}


	//! Check to see if a given property is set on this cell.
	//!
	//! This is called IsPropertySet() on the Ultimate Grid Control.
	bool HasProperty (
		PROPERTY property
		) const {
		return ((m_set & property) != PROP_None);
		}

	//! Set the alignment.
	void SetAlignment (
		ALIGNMENT alignment
		);

	//! Set background color.
	void SetBackColor (
		const COLOR& color
		);

	//! Set the backgound color and make hilighted. 
	void SetBackColorHilighted (
		const COLOR& color
		);

	//! Sets a boolean value.
	//!
	//! For toggle button type cells, toggles the cell on/off.
	void SetBool (
		bool state
		);

	//! Set the border.
	void SetBorder (
		BORDER border
		);

	//! Set the cell to the given cell type.
	//!
	//! These are not enums like the SetDataType.  
	//! Cell types are the result of calling MIGRIDCTRL::AddCellType()
	void SetCellType (
		CELLTYPE type
		);

	//! Sets a cell "subtype", the value of which depends on the type of cell.
	void SetCellTypeEx (
		INT32 val
		);

	//! Set the data type (text, number, bool, etc) for the cell.
	void SetDataType (
		DATATYPE type
		);

	//! Set the foreground color.
	void SetForeColor (
		const COLOR& color
		);

	//! Set the foreground color and make hilighted.
	void SetForeColorHilighted (
		const COLOR& color
		);

	//! Set the index of the icon associated with this entry.
	//! 
	//! Sets the index into the IMAGELIST for this grid.  The IMAGELIST
	//! Is set with the MIGRIDCTRL::SetImageList() method.
	//! Setting this to -1 means no icon.
	//! If an item has an icon, it will be displayed to the left of its
	//! content.
	void SetIconIndex (
		int icon
		);

	//! Set the "label".
	//!
	//! How this string is used is extremely celltype
	//! dependant.  The name comes from the UltimateGrid control.
	//! It is ignored for most cell types.
	//!
	//!	CELLTYPE_Dropdown
	//!		will use it as a list of values for the dropdown list.  You 
	//!		seperate the possible values with "\n"s.  Be sure that the 
	//!		last one ends in a \n or it won't show up 
	//!
	//! CELLTYPE_Icon 
	//!		Set to the name of the icon to show in the button/label.  The
	//!		string should be of the form  "group.name" (since there's only
	//!		one string and you need to be able to set two parts)
	//!	
	//!	CELLTYPE_IconButton
	//!		Same as CELLTYPE_Icon, but optionally you can give it two icon
	//!		names such as "group.officon group.onicon".  The second entry
	//!		will be used when the button is in the on state.  To make a
	//!		button act as a toggle, set the CellTypeEx(CELLTYPEEX_Toggle)
	//!		and use SetBool() to turn the button on/off
	void SetLabelText (
		const MIUNICODE*
		);

	//! Since this is used for naming icons, and we'd like to pass simple hardcoded strings.  
	//!
	//! For populating lists, you really want to use the Unicode version above.
	void SetLabelText (
		const char*
		);

	//! Set the maximum lenght of the field.  
	//!
	//! Setting to -1 will make it no maximum.
	void SetMaxLen (
		int num
		);

	//! Sets the value of the cell as a number. 
	//!
	//! Setting a cell to IEEE_NaN will cause it to be "empty".
	void SetNumber (
		double num
		);

	//! Set the number of decimal places.
	void SetNumberDecimals (
		int num
		);

	//! Sets a general-usage UserData pointer which can be retrieved with GetParam.
	void SetParam (
		void* ptr
		);

	//! Set to read only.
	void SetReadOnly (
		bool state
		);

	//! Set the text that will show in the cell.
	void SetText (
		const MIUNICODE*
		);

	//! Merges the properties from a given cell into this cell.
	//!
	//! Any properties already set in this cell are left alone (except 
	//! colors which can be blended if set in both cells).
	void MergeProperties (
		const MIGRIDCELL& cell,
		bool bBlendColors = false,
		PROPERTY mask = PROP_All
		);

private:
	#ifndef GENERATING_DOXYGEN_OUTPUT
	
	// METHODS

	void SetProperty (
		PROPERTY prop
		) {
		m_set = static_cast<PROPERTY>((UINT32)prop | (UINT32)m_set);
		}

	//////////////////////////////////////////////////////////////////////////////!
	//! Private Members
	//! The size of one of these is starting to get a bit large.  I may change this
	//! so that the data portion is in a base class and MIGRIDCELL is a derived
	//! class with all the style info.  Internally, the grid would only store
	//! the base class if the cell has no styles of its own.  Of course, you can
	//! implement your whole grid by filling the data in OnGetCell() and never
	//! calling SetCell(), in which case it never has to store anything internally.

	PROPERTY m_set;		//!< Bits of which properties have been set
	MIUNICODE* m_str;
	MIUNICODE* m_Label;
	COLOR m_BGColor;
	COLOR m_FGColor;
	COLOR m_BGHColor;		//!< Background color while highlighted
	COLOR m_FGHColor;		//!< Foreground color while highlighted
	BORDER m_BorderStyle;
	ALIGNMENT m_Alignment;
	DATATYPE m_DataType;
	CELLTYPE m_CellType;
	INT32 m_nCellTypeEx;
	bool m_bReadOnly;
	bool m_bState;			//!< For boolean fields only
	int m_nPlaces;
	void* m_pParam;
	int m_nMaxLen;
	int m_IconIndex;

	static const COLOR s_black;	//!< Things that return (const COLOR&) return this if nothing to return
	#endif // GENERATING_DOXYGEN_OUTPUT
	};


#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MIGRIDCELL::BORDER)
DEFINE_ENUM_OP_BITWISE(MIGRIDCELL::PROPERTY)
#endif


//////////////////////////////////////////////////////////////////////////////!
//////////////////////////////////////////////////////////////////////////////!
//! A structure passed to MIGIRDCTRL::OnDrawItem()
struct MIGRIDCUSTOMDRAW {
	//! Constructor initializes the structure 
	MIGRIDCUSTOMDRAW (
		const MIGRIDCELL&, 
		const LRECT2D&, 
		MIGRIDCTRL&,
		MGD::CONTEXT* 
		);

	MIGRIDCUSTOMDRAW (
		const MIGRIDCUSTOMDRAW& rhs
		);

	LRECT2D rect;
	COLOR BackgroundColor;
	COLOR TextColor;
	const MIGRIDCELL& cell;
	bool ChangedBackgroundColor;	//!< Set to true if you changed BackgroundColor
	bool ChangedTextColor;	//!< Set to true if you changed TextColor
	MGD::CONTEXT* gc;
	bool bSelected;
	bool bHasFocus;
	MIGRIDCTRL* pCtrl;
	};


//////////////////////////////////////////////////////////////////////////////!
//////////////////////////!                          //////////////////////////
/////////////////////!           MIGRIDCELLTYPE          //////////////////////
//////////////////////////!                          //////////////////////////
//////////////////////////////////////////////////////////////////////////////!

//!	An interface only base class for various types of data.
//! This could go into an migridp.h
class MIGRIDCELLTYPE {
public:

// CONSTRUCTION / DESTRUCTION

	//! Constructor.
	MIGRIDCELLTYPE (
		) :
		m_bDeleteWithGrid(false)
		{
		}

	//! Destructor.
	virtual ~MIGRIDCELLTYPE (
		) {
		}

	// METHODS

	
	//! Default method just draws an empty cell with background and borders.
	virtual void v_Draw (
		int row,
		int col,
		MIGRIDCUSTOMDRAW& dd
		); 

 	//! Function to call to start editing.
 	//! 
 	//! To cancel the edit, delete the handle.  
 	//! Default method just returns NULL and doesn't allow editing.
	virtual MIGRIDEDITHANDLE* v_StartEdit (
		MIGRIDCTRL* pCtrl,
		int col,
		int row,
		const LRECT2D& rect,
		const LPOINT2D* point,				//!< Click point within cell (NULL if not starting via click)
		const MIGRIDCELL& cell,
		int key = 0
		);

	//! Method called when a cell is double clicked on.
	//!
	//! It's important to set *pbProcessed = true if you actually handle the double 
	//! click. If you don't it will be treated as just another left click (and call OnLClicked())
	//! Default method does nothing.
	virtual void v_OnDClicked (
		int col,
		int row,
		const LRECT2D& rect,
		const LPOINT2D& point,
		bool* pbProcessed,
		MIGRIDCTRL* pCtrl,
		UINT32 state		//!< ShiftMask, CtrlMask, etc from the XEvent
		);

	//! Default method does nothing.
	virtual void v_OnLClicked (
		int col,
		int row,
		bool bIsDown,
		const LRECT2D& rect,
		const LPOINT2D& point,				//!< Point relative to rect
		bool* pbProcessed,
		MIGRIDCTRL* pCtrl,
		UINT32 state		//!< ShiftMask, CtrlMask, etc from the XEvent
		);

	//! Default method does nothing.
	virtual void v_OnRClicked (
		int col,
		int row,
		bool bIsDown,
		const LRECT2D& rect,
		const LPOINT2D& point,
		bool* pbProcessed,
		MIGRIDCTRL* pCtrl,
		UINT32 state		//!< ShiftMask, CtrlMask, etc from the XEvent
		);

	//! Get whether delete with grid value.
	bool GetDeleteWithGrid (
		) const {
		return (m_bDeleteWithGrid);
		}

	//! Set the delete with grid value.	
	void SetDeleteWithGrid (
		bool state
		) {
		m_bDeleteWithGrid = state;
		}

	//! Called from MIGRIDCTRL::Create() giving us the default pixel values used by Motif.  
	static void SetDefaultColors (
		MGD::PIXEL ShadowPixelLight,
		MGD::PIXEL ShadowPixelDark,
		MGD::PIXEL FocusRectPixel,
		MGD::PIXEL BackgroundPixel
		) {
		s_ShadowPixelLight = ShadowPixelLight;
		s_ShadowPixelDark = ShadowPixelDark;
		s_FocusRectPixel = FocusRectPixel;
		s_BackgroundPixel = BackgroundPixel;
		}

protected:
	//! Default method just draws background normally derived class could override this necessary.
	virtual void v_DrawBackground (
		const MIGRIDCUSTOMDRAW&  drawdata
		);

	//! Default method just draws border normally derived class could override this necessary.
	virtual void v_DrawBorder (
		const MIGRIDCUSTOMDRAW&  drawdata
		);

	//! It just draws whatever type of border you tell it to with the given rect.
	//!
	//! This version is not overridable.  
	void DrawBorder (
		const MIGRIDCUSTOMDRAW&  drawdata,
		MIGRIDCELL::BORDER border
		);

	//! Draws an empty button which can be tracked via TrackButton.
	//!
	//! Default method just draws button border with nothing in it.
	virtual void v_DrawButton (
		const MIGRIDCUSTOMDRAW&  drawdata,
		bool bPressed,
		bool bTracking	= false				//!< True only between mouse down and mouse up
		);

	//! Default method just draws text normally. derived class could override 
	//! this if necessary.
	//! If the str parameter is left NULL (the default), it will get the text
	//! from the cell in the drawdata.  If not null, it overrides the cell text
	virtual void v_DrawText (
		const MIGRIDCUSTOMDRAW&  drawdata,
		const MIUNICODE* str = 0
		);

	//! Track the button.
	void TrackButton (
		MIGRIDCTRL* pCtrl,
		int col,
		int row,
		const MIGRIDCELL& cell,
		const LRECT2D& rect,
		const LPOINT2D& point,
		bool bInitialState
		);

	//! Called by TrackButton when the button is actually pressed.
	//!
	//! Default method calls pCtrl->StartEdit(col, row, 0).
	virtual void v_OnButtonPress (
		MIGRIDCTRL* pCtrl,
		int col,
		int row,
		const MIGRIDCELL& cell,
		const LRECT2D& rect
		);

	//! Default pixel values to be used if cell being drawn has no color specified
	//! These are set via a static method called from MIGRIDCTRL::Create(), which
	//! gives us default values from X.
	static MGD::PIXEL s_ShadowPixelLight;
	static MGD::PIXEL s_ShadowPixelDark;
	static MGD::PIXEL s_FocusRectPixel;
	static MGD::PIXEL s_BackgroundPixel;

private:
	#ifndef GENERATING_DOXYGEN_OUTPUT
#ifdef X_NATIVE
	static void EH_TrackButtonEventHndler (
		Widget da, 
		void *ehdata, 
		XEvent *event,
		Boolean *ContToDispatch		//!<  Set to false to slurp the event 
		);
#endif

	bool m_bDeleteWithGrid;		//!< If true, grid's destructor will delete this celltype
	friend class MIGRIDCTRL;
	#endif // GENERATING_DOXYGEN_OUTPUT
	};

//////////////////////////////////////////////////////////////////////////////!
//////////////////////////!                          //////////////////////////
////////////////////!           MIGRIDEDITHANDLE          /////////////////////
//////////////////////////!                          //////////////////////////
//////////////////////////////////////////////////////////////////////////////!

//!
//!	The various MIGRIDCELLTYPE return one of these from their StartEdit()
//!	method.  This could go into an migridp.h
//! This class is derived from MIGRIDCELL since it is closely tied to
//! the cell that's being edited.  (It actually keeps a COPY of the cell
//! being edited)
//!
class MIGRIDEDITHANDLE : public MIGRIDCELL {
public:

// CONSTRUCTION / DESTRUCTION

	//! Constructor.
	MIGRIDEDITHANDLE (
		MIGRIDCTRL* pCtrl,
		int col,
		int row,
		const LRECT2D& rect,
		const MIGRIDCELL& cell
		) :
		MIGRIDCELL(cell),
		m_pCtrl(pCtrl),
		m_rect(rect),
		m_col(col),
		m_row(row)
		{
		}

	//! Destructor defined in migrdcel.c.
	//!
	//! Calls MIGRIDCTRL::EditFinish() passing this and true,
	//! i.e., if we destroy the handle before we're done editing, cancel
	virtual ~MIGRIDEDITHANDLE (
		);

	// METHODS
						
	//! Retrieves the XeText widget or CWnd of the Text Edit control if it has one.  
	//!
	//!	@return The XeText widget or CWnd of the Text Edit control if it has one.
	//!
	//! Used to get the XeText to pass back to v_OnEditFinish()
	MDLGPARENT GetTextControl (
		) const { return v_GetTextControl(); }

	//! Update() should make sure that the data in the handle reflects whatever GUI elements it has.
	//! 
	//! @return True if value changed from previous value, false if no change.
	bool Update (
		) { return v_Update(); }

	//! This method is called whenever the user scrolls to move the editable area.
	void Move (
		const LRECT2D&  rect
		) { v_Move(rect); }
	
protected:

	//! Retrieves the XeText widget or CWnd of the Text Edit control if it has one.  
	//!
	//!	@return The XeText widget or CWnd of the Text Edit control if it has one.
	//!
	//! Used to get the XeText to pass back to OnEditFinish()
	virtual MDLGPARENT v_GetTextControl (
		) const {
		return NULL;
		}

	//! This method is called whenever the user scrolls to move the editable area.
	virtual void v_Move (
		const LRECT2D& // rect
		) {
		}

	//! Update() should make sure that the data in the handle reflects whatever GUI elements it has.
	//! 
	//! @return True if value changed from previous value, false if no change.
	virtual bool v_Update (
		) {
		return true;
		}

	//! Send that a mouse event occurred.
	void SendMouseEvent (
		Widget widget,
		int EventCode,
		const LPOINT2D* point
		);

	MIGRIDCTRL* m_pCtrl;
	LRECT2D m_rect;
	int m_col;
	int m_row;
	};


//////////////////////////////////////////////////////////////////////////////!
//////////////////////////!                          //////////////////////////
///////////////////////!           MIGRIDCTRL          ////////////////////////
//////////////////////////!                          //////////////////////////
//////////////////////////////////////////////////////////////////////////////!
//!
//!	The MIGRIDCTRL API is based on the UltimateGrid control from Dundas software.
//!	MIGRIDCTRL tries to emulate as much of the UltimateGrid as possible.  One 
//!	of the design goals for writing this was to make it possible to write X
//!	code using MIGRIDCTRL and later port it to MFC by substituting the control
//!	from Dundas. Some of the control creation will have to change, but a lot
//!	of it will be the same.  (This is why some of the methods don't follow
//!	standard MicroImages naming standards and abbreviations)

class MIGRIDCTRL {
	public:

		enum SECTION {
			SECTION_None	= 0,
			SECTION_Grid	=1,
			SECTION_TopHeading,
			SECTION_SideHeading,
			SECTION_CornerButton,
			SECTION_HScroll,
			SECTION_VScroll,
			SECTION_Tab,
			SECTION_ColBorder,
			SECTION_RowBorder
			};

		enum SCROLLMODE {
			SCROLLMODE_Normal,				//!<	Doesn't update while draging, just on release
			SCROLLMODE_Tracking				//!< Updates grid while draging slider
			};

		//! States for OnCellTypeNotify.
		enum CELLSTATE {
			CELLSTATE_None		= 0x0000,	//!< No state falgs set
			CELLSTATE_Button	= 0x0001,	//!< Button (GUI componant) is "on"
			CELLSTATE_Shift	= 0x0002,	//!<	Shift key is held down
			CELLSTATE_Control	= 0x0004		//!< Control key is held down
			};

		//! Values for MoveCurrentRow/MoveCurrentCol.
		enum MOVEMENT {
			MOVE_Increment = 0,
			MOVE_Decrement,
			MOVE_PageUp,
			MOVE_PageDown,
			MOVE_Top,
			MOVE_Bottom
			};
			
		struct SORTDATA {
			MIGRIDCTRL* m_this;
			int* m_cols;
			int m_NumCols;
			int m_bAscending;
			};

		//! A function for testing only.  Will go away later.
		static void Test (
			DBEDIT* ddata,
			DBEDIT_TDATA* tdata
			);

		// CONSTRUCTION / DESTRUCTION

		//! Standard constructor.
		MIGRIDCTRL (
			);

		// Copy constructor goes here?

		//! Destructor.
		virtual ~MIGRIDCTRL (
			);

		// METHODS

		//! Adds a cell type.  
		//!
		//! @return A celltype index which can be used in calls to MIGRIDCELL::SetCellType().
		//!
		//! The bDeleteInGridsDestructor should be set to true if you're doing...
		//!		grid->AddCellType(new MIGRIDCELLTYPE_SOMETHING, true);
		//! to let the grid know it should delete the celltype in its destructor.
		MIGRIDCELL::CELLTYPE AddCellType (
			MIGRIDCELLTYPE* ptype,
			bool bDeleteInGridsDestructor = false
			);

		//! This is like InsertCol() but will add the column to the end.
		//!
		//! NOTE: Much more efficient of done while there are no rows.
		int AppendCol (
			);

		//! This is like InsertRow() but will add the row to the end.
		int AppendRow (
			);

		//! Unselect all cells.
		int ClearSelections (
			);

#ifdef X_NATIVE
		int Create (
			Widget parent,
			Arg *arg,
			int n
			);
#endif

		void DeleteAllItems (
			) {
			// Used by MGUI::LIST
			SetNumberRows(0);
			}
		
		//! Delete a column.
		int DeleteCol (
			int col
			);

		void DeleteItem (		
			int row
			) {
			// Used by MGUI::LIST
			DeleteRow(row);
			}

		//! Delete a row.
		int DeleteRow (
			int row
			);

		//! Will stop current edit (if active) without saving changes.
		int EditCancel (
			);

		//! Finish the editing.
		//! 
		//! This function is called internally (but by a class which isn't 
		//! a friend, or I'd make this method private)
		//! Unless aborting is set, it will call OnEditFinish, to verify
		//! the data.  Will then, if necessary, retain the values
		//! internally and redraw the cell.
		void EditFinish (
			MIGRIDEDITHANDLE* handle = NULL,
			bool bAborting = false
			);

		//! Finish the editing.
		void EditFinish (
			int col,
			int row,
			const MIGRIDCELL& cell
			);

		//! Allow user to swap columns by dragging them.
		int EnableColSwapping (
			bool bEnable = true
			);

		//! Will enable/disable the right-mouse button popup menu.
		int EnableMenu (
			bool bEnable = true
			);

		//! Enable/Disable updates.  
		//!
		//! Keeps track of how many times it's been disabled.  When reeneabled, will automatically
		//! redraw if needed.
		void EnableUpdate (
			bool enable = true
			);

		//! Make sure that a given row and column are visible
		//! Pass -1 for either row or column to only scroll in the other direction
		void EnsureVisible (
			int row,
			int col,
			bool bPartialOK = false
			);

		int FindItemByData (
			PTRINT data
			) const {
			return (FindItemByData((const void*)data));
			}

		//! Returns the row number in which the Param value for column 0
		//! is equal to the given value.  Returns -1 if no row with that value.
		int FindItemByData (
			const void* data
			) const;

		//! Given an x and y, return a row and column number.  
		//! @return 0 for success, -1 if no cell at given point.
		int GetAbsoluteCellFromPoint (
			int x,								//!< X coordinate relative to upper left of grid control
			int y,								//!< Y coordinate relative to upper left of grid control
			int* ptcol,							//!< column number returned
			int* ptrow  						//!< row number returned
			) const;

		//! Retrieves the row number of the bottom most visable row.
		//!
		//! @return The row number of the bottom most visable row.
		int GetBottomRow (
			) const;

		//! Gets a COPY of a given cell.  
		//!
		//! @return Only properties which have been directly set for the cell
		//! Normally, you get the cell, modify it using its methods, then call SetCell.
		//! Returns 0 if Cell exists, -1 if not, < -1 if error.
		int GetCell (
			int col,
			int row,
			MIGRIDCELL* pCell
			) const;


		//! This version sets *ppcell to a pointer to the internal cell data IF THERE IS ANY and returns 0.  
		//! 
		//! If there is no internal data for
		//!	the cell, *ppCell will be NULL and the function will return -1
		int GetCell (
			int col,
			int row,
			MIGRIDCELL** ppCell
			) const;

		//! Just like GetCell, but merges the cells properties with the default properties of the row and colum which it is in.
		//! 
		//! Values are retrieved in the following order.
		//!		1. Column properties
		//!		2. Row properties
		//!		3. Actual cell contents
		//!		4. Calls OnGetCell() which can override values.  
		//!			changes made by OnGetCell() only affect this
		//!			copy, not the unerlying data.
		int GetCellIndirect (
			int col,
			int row,
			MIGRIDCELL* pCell,
			MIGRIDCELL::PROPERTY properties = MIGRIDCELL::PROP_All
			);	//! No, not const.  Calls OnGetCell() and may cache result
			
		//! Get the rectangle of a cell in coordinates relative to the upper left corner of the drawing area it's in.
		int GetCellRect (
			int col,
			int row,
			LRECT2D* rect
			) const;

		//! Retrieves a cell with the default settings for a column (background color, alignment, etc). 
		//! 
		//! @return A cell with the default settings for a column.
		//!
		//! NOTE: if there is no default for the column, this will
		//! return a cell with nothing set, not the grid default
		int GetColDefault (
			int col,
			MIGRIDCELL* pCell
			) const;

		//! Retrieves the logical column number of then n'th column in visual order.
		//!
		//! @return The logical column number of then n'th column in visual order.
		//!
		//! Note: Assumed that columns can be swapped around visually.
		int GetColTranslation (
			int VisualColumnNumber
			) const;


		//! Retrieves the column order as an array.
		//! Note: Assumed that columns can be swapped around visually.
		ERRVALUE GetColumnOrderArray (
			SIMPLE_ARRAY<INT32>&
			) const;

		//! Get the width of a column in pixels.
		//!
		//! @return 0 for success, -1 for error.
		int GetColWidth (
			int col,
			int* width
			) const;

		//! Get the width of a column in pixels.	
		//! 
		//! @return The width, 0 if the column number is invalid.
		int GetColWidth (
			int col
			) const;

		//! Get the width of a column in pixels.
		//! 
		//! @return The width, 0 if the column number is invalid.
		int GetColumnWidth (
			int col
			) const {
			return (GetColWidth(col));
			}

		//! Get the number of rows complete rows that can fit on the page.
		int GetCountPerPage (
			) const;

		//! Retrieves the column of the cell which has focus (or would have focus if the grid had focus).
		//!
		//! @return The column of the cell which has focus.
		int GetCurrentCol (
			) const;

		//! Retrieves the row of the cell which has focus (or would have focus if the grid had focus).
		//! 
		//! @return The row of the cell which has focus.
		int GetCurrentRow (
			) const;

		//! Retrieves the number of the current tab if the control has more than one.  
		//!
		//! @return The number of the current tab if the control has more than one.
		//!
		//! Note: I may or may not implement this.
		int GetCurrentTab (
			) const {
			return 0;
			}

#ifdef X_NATIVE
		//! Get the main XmForm of the control.
		//! The main form can be used to control attachment to other widgets.
		//! Note: This is not available in the WIN32_NATIVE build
		Widget GetForm() const {
			return (m_wMainForm);
			}
#endif

		//! Returns a cell with the default style for the grid.
		//!
		//! @return A cell with the default style for the grid.
		int GetGridDefault (
			MIGRIDCELL* pCell
			) const;

		//! Returns a cell with the default style for the heading.
		//!
		//! @return A cell with the default style for the heading.
		int GetHeadingDefault (
			MIGRIDCELL* pCell
			) const;

		//! Returns the height of the horizontal scrollbar in pixels.
		//!
		//! @return The height of the horizontal scrollbar in pixels.
		int GetHSHeight (
			) const;


		//! Return the ImageList associated with this Grid.
		//! Return value will be NULL if no image list associated with the grid.
		//! Call SetImageList to get the image list.
		const MGUI::IMAGELIST* GetImageList() const {
			return (m_ImageList);
			}

		//! Get the userdata (PTRUINT) for an item in the list.
		PTRUINT GetItemData (
			int row,
			int col = 0
			) const {
			return (reinterpret_cast<PTRUINT>(GetItemDataPtr(row, col)));
			}

		//! Get the userdata (void*) for an item in the list.
		void* GetItemDataPtr (
			int row,
			int col = 0
			) const;

		//! Returns the column number of the left-most visable column, counting partial columns.
		//!
		//! @return The column number of the left-most visable column, counting partial columns.  
		//!
		//! NOTE:  If column swapping is enabled,
		//! don't assume that columns go sequentially from left to right.
		//! In fact.  This is a pretty useless value.
		int GetLeftCol (
			) const;

		//! Returns the row number of the next row with a selection in it.
		//! Returns -1 if no more selected items
		//! Pass -1 to start at the top.  The row number passed is ignored.
		//! Note: This is horribly inefficient.  GetSelectedItemList() is
		//! more efficient.
		int GetNextSelectedItem (
			int row
			) const;
			
		//! Returns the number of columns in the grid.
		//!
		//! @return The number of columns in the grid.
		int GetNumberCols (
			) const;

		//! Returns the number of rows in the grid.
		//!
		//! @return The number of rows in the grid.
		int GetNumberRows (
			) const;

		int GetItemCount (
			) const {
			return (GetNumberRows());
			}

		//! Returns the column number of the right-most visable column, counting partial columns.
		//!
		//!	@return The column number of the right-most visable column, counting partial columns.
		//! 
		//! Actually, this method may go away.
		//! Since columns can be swapped, knowing which one is on the 
		//! right is pretty useful.
		int GetRightCol (
			) const;

		//! Returns the height of a given row.
		//!
		//! @return The height of a given row.
		//!
		//! Note that not all rows have to be the same height.
		int GetRowHeight (
			int row
			) const;

		int GetSelectedCount() const {
			return (m_SelectedCells.GetNumItems());
			}

		//! Returns an array of selected cells.
		//! For each cell, the y value is the row, the x value is the column.
		//! An x of -1 indicates the whole row is selected.
		const SIMPLE_ARRAY<LPOINT2D>& GetSelectedItems (
			) const {
			return (m_SelectedCells);
			}

		//! Returns the width of the side heading, which does not scroll horizontally with the rest of the grid.
		//!
		//! @return The width of the side heading, which does not scroll horizontally with the rest of the grid.
		int GetSideHeadingWidth (
			) const;

		//! Gets the size of the main form in pixels.
		//!
		//! Size returned includes headings and scroll bars.
		void GetSize (
			LPOINT2D& size
			) const;

		//! Returns the height of the top heading, which does not scroll vertically with the rest of the grid.
		//!
		//! @return The height of the top heading, which does not scroll vertically with the rest of the grid
		int GetTopHeadingHeight (
			) const;
	
		//! Returns the index of the top-most (partially) visable row.
		//!
		//! @return The index of the top-most (partially) visable row.
		int GetTopRow (
			) const;

		//! Returns the width of the vertical scrollbar.
		//!
		//! @return The width of the vertical scrollbar.
		int GetVSWidth (
			) const;

		//! Change the focus to the given cell.  
		//!
		//! @return 0 on success, -1 the cell is invalid or if OnCanMove() disallowed the move.
		int GotoCell (
			int col,
			int row
			);

		//! Change the focus to the given cell.
		//!
		//! @return 0 on success, -1 the cell is invalid or if OnCanMove() disallowed the move.
		//!
		//! Changes only the column, leaving the row the same.
		int GotoCol (
			int col
			) {
			return (GotoCell(col, GetCurrentRow()));
			}

		//! Change the focus to the given cell.
		//!
		//! @return 0 on success, -1 the cell is invalid or if OnCanMove() disallowed the move.
		//!
		//! Changes only the row, leaving the column the same.  
		int GotoRow (
			int row
			) {
			return (GotoCell(GetCurrentCol(), row));
			}	

		//! Sets the current cell to be -1-1, thereby removing the focus rectangle.
		int HideCurrentCell (
			) {
			return (GotoCell(-1,-1));
			}

		//! Insert a column before the specified column.
		//!
		//! NOTE: Much more effeciant of done while there are no rows
		int InsertCol (
			int col
			);

		//! Insert a column before the specified column.
		int InsertColumn (
			int col,
			const MIUNICODE* string,
			MIGRIDCELL::ALIGNMENT alignment,
			int width = -1
			);

		//! Insert a row before the specified row.
		int InsertRow (
			int row
			);

		//! Find if cell is selected.
		//!
		//! @return True if the given cell is selected.
		bool IsSelected (
			int col,
			int row
			) const;

		//! Locks or freezes the specified number of columns so that they will not scroll off the screen.
		//!
		//! Locked columns start from column 0, so
		//! calling this function with a value of 3 will lock columns 0, 1, and 2.
		//! To unlock the columns, call LockColumsn(0).
		//! Not implemented yet.
		int LockColumns (
			int cols
			);

		//! Locks or freezes the specified number of rows so that they will not scroll off the screen.
		//!
		//! Locked rows start from rows 0, so
		//! calling this function with a value of 3 will lock rows 0, 1, and 2.
		//! To unlock the rows, call LockRows(0).
		//! Not implemented yet.
		int LockRows (
			int rows
			);

		//! Function used to reorder columms visually.
		//!
		//! Column numbers are logical, not visual.  
		int MoveColPosition (
			int fromCol,
			int toCol,
			bool insertBefore
			);

		//! Used to move the focus rectangle relative to current focus rectangle.
		//!
		//! @return 0 on success, -1 if movement can't happen.
		//!
		//! MOVE_Increment, MOVE_PageDown, MOVE_Bottom all go to the right.
		int MoveCurrentCol (
			MOVEMENT movement
			);

		//! Used to move the focus rectangle relative to current focus rectangle.
		//! 
		//! @returns 0 on success, -1 if movement can't happen.
		int MoveCurrentRow (
			MOVEMENT movement
			);

		//! Called to determine if column swapping should be allowed.
		//!
		//! @return True to allow, false to disallow.
		//!
		//! In the case of visually dragging a column, we're not really
		//! swapping the columns (This is the UltimateGridControls's 
		//! method name, not mine).  Column numbers are real logical
		//! column numbers, not visual.
		virtual bool v_OnCanColSwap (
			int fromCol,						//!< column we're moving
			int toCol							//!< column we're about to place it AFTER (-1 if making fromCol the first column)
			);

		//! Called to determine if a given change of cell focus should be allowed.
		//!
		//! @return False to prevent, true to allow.
		virtual bool v_OnCanMove (
			int oldcol,
			int oldrow,
			int newcol, 
			int newrow
			);

		//! Called to determine if user can resize a column.
		virtual bool v_OnCanSizeCol (
			int col
			);

		//! Called to determine if user can resize a row (row resizing not implemented anyway).
		virtual bool v_OnCanSizeRow (
			int row
			);

		//! Called to determine if user can resize the side heading (heading resizing not implemented anyway).
		virtual bool v_OnCanSizeSideHdg (
			);

		//! Called to determine if user can resize the top heading (the height) (heading resizing not implemented anyway).
		virtual bool v_OnCanSizeTopHdg (
			);

		//! Called to notify you that the cell focus has changed.
		virtual void v_OnCellChange (
			int oldcol,
			int oldrow,
			int newcol,
			int newrow
			);

		//! Called when certain cell type actions are performed.
		//!  
		//! @return False to prevent the action from happening (but it's already happened, so why bother?).
		//!
		//! See the comments above where MIGRIDCELL::CELLTYPE is defined for details.
		virtual bool v_OnCellTypeNotify (
			MIGRIDCELL::CELLTYPE celltype,
			int col,
			int row, 
			INT32 msg,
			CELLSTATE param	
			);

		//! Called to notify you that the cell focus has changed.
		virtual void v_OnColChange (
			int oldcol,
			int newcol
			);

		//! Called to inform you that the width of a column is being changed.  
		//!
		//! Note that the width is passed by reference and
		//! CAN be ammended.  This way, you can implement min/max sizes.
		//! Note, this function is called by SetColumnWidth().
		virtual void v_OnColSized (
			int col,
			int* width
			);

		//! Called to inform you that the width of a column is being changed. 
		//!
		//! Note that the width is passed by reference and
		//! CAN be ammended.  This way, you can implement min/max sizes.
		//! Note, this function is called while the resize action is
		//! taking place (user is dragging the divider between columns).
		virtual void v_OnColSizing (
			int col,
			int* width
			);

		//! Called when the user double clicks on a cell.
		//!
		//! The point passed in is coordinates relative to the upper
		//! left of the cell where the click took place.
		//! The rect is the extents of said cell (relative to the drawing 
		//! area it's drawing in).
		//! NOTE:  If you use the event, set *pbProcessed to true
		//! to prevent the event from propagating.  If you don't
		//! it will be treated as just another left click (and call 
		//!	OnLClicked()).
		virtual void v_OnDClicked (
			int col,
			int row,
			LRECT2D* rect,
			LPOINT2D* point,
			bool* pbProcessed,
			UINT32 state		//!< ShiftMask, CtrlMask, etc from the XEvent
			);

		//! Called before an item is drawn.  Override if you want to do the 
		//! drawing yourself.  Return true to prevent the list from drawing,
		//! false to let it draw. 
		virtual bool v_OnDrawItem (
			int row,
			int col,
			MIGRIDCUSTOMDRAW& drawdata
			);

		//! Called while the cell is still in the editing mode.
		//!
		//! @return True to allow changes, false to reject.
		//!
		//! This allows you to cancel any changes, or prevent the change of focus. 
		//! I may not implemnt this one, as I don't see how it differs
		//! from OnEditFinish and OnCellChange.  (Not implemented)
		virtual bool v_OnEditContinue (
			int oldcol,
			int oldrow,
			int* newcol,
			int* newrow
			);
			
		//! Called when a cell has been edited.  
		//! 
		//! @return False to reject the changes, true to accept them.
		//!
		//! I'd rather make this get the MIGRIDCELL from which
		//! you could get the string or bool value, but UltimateGrid control
		//! gives you the string, so that's all I got.   For the X version,
		//! I'll make the string equal "1" or "0" for boolean cell types.
		//! Note that TextControl will be NULL unless it's a text cell.
		virtual bool v_OnEditFinish (
			int col,
			int row,
			MDLGPARENT TextControl,
			const MIUNICODE* string
			);

		//! Called when a cell has been edited.  
		//! 
		//! @return False to reject the changes, true to accept them.
		virtual bool v_OnEditFinish (
			int col,
			int row,
			const MIGRIDCELL& cell
			);

		//! Called when a cell is about to be edited.  
		//!
		//! @return return false to prevent edit, true to allow
		//!
		//! In X version, this is only called when editing begins on a
		//!	text cell (buttons and dropdowns don't call it).
		virtual bool v_OnEditStart (
			int col,
			int row
			);

		//! Called before adding text to a string.
		//!
		//! @return True to allow it, false to reject it.
		//!
		//! UGC also gets CEdit* of the edit control 
		//! X version is only passed the text about to be inserted.
		//! (I don't know what UGC passes, documentation is unclear)
		//! WARNING!  If the edit is a deletion, str may be NULL.
		//! Hmmm!  In the UGC version, "str" is a pointer to a UINT
		//! holding a single char which may not by Unicode.
		//!	This also gets VK_DELETE, VK_TAB, etc.
		virtual bool v_OnEditVerify (
			int col,
			int row,
			MDLGPARENT TextControl,
			const MIUNICODE* str
			);

		//! Called by GetCellIndirect() to let derived class modify cell properties and even the cell value.  
		//!
		//! Note that if the data has been sorted by calling SortBy(), you should
		//! not rely on the row number being the same order that
		//! you added the rows.
		virtual void v_OnGetCell (
			int col,
			int row,
			MIGRIDCELL* cell,
			MIGRIDCELL::PROPERTY mask = MIGRIDCELL::PROP_All
			);


		//! Called when the mouse has been over a cell long enough to produce a tooltip.  
		//!
		//! @return True to show hint, false to not.
		//!
		//! Call SetText or SetImage on the tooltip control passed in.
		//! The default implementation calls the MIUNICODE version
		//! NOTE:  You have to call UseHints() to enable this.
		virtual bool v_OnHint (
			int col,
			int row,
			SECTION section,
			MGUI::CTRL_TOOLTIP& tooltip
			);

		//! Called while horizontal scrolling to get the tooltip to show.
		//!
		//! NOTE:  You have to call UseHScrollHints() to enable this.
		virtual bool v_OnHScrollHint (
			int col,
			MISTRING&
			);
			
		//! Called by SortBy() to compare items.
		//! Should return < 0 if item1 should precede item2 and > 0 if item1 should
		//! follow item2, and 0 if they're equivalent.
		//! The default implementation calls GetCellIndirect() and compares the
		//! cell values.
		virtual int v_OnItemCompare (
			UINT32 item1,
			UINT32 item2,
			const SORTDATA* sortdata
			);
		
		//! Called when an item is selected.  
		//! Return true to allow it to be selected, false to prevent selection.
		virtual bool v_OnItemSelect (
			int row,
			int col
			);

		//! Called when an item is unselected.  
		//! Return true to allow it to be unselected, false to prevent unselection.
		//! Note that a call to ClearSelections will call OnItemUnselect() but
		//! will ignore the return value.  
		virtual bool v_OnItemUnselect (
			int row,
			int col
			);

		//! Called when focus is lost for the whole grid.
		//!
		//! [ XXX Not implemented yet ]
		virtual void v_OnKillFocus (
			SECTION section
			);
		
		//! Called when the user left clicks on a cell.
		//!
		//! The point passed in is coordinates relative to the upper left of the cell where 
		//! the click took place.  The rect is the extents of said cell.
		//! NOTE:  If you use the event, set *pbProcessed to true
		//! to prevent the event from propagating.
		virtual void v_OnLClicked (
			int col,
			int row,
			bool bIsMouseDown,
			LRECT2D* rect,
			LPOINT2D* point,
			bool* pbProcessed,
			UINT32 state		//!< ShiftMask, CtrlMask, etc from the XEvent
			);

		//! Called when the left-most visable column changes.
		virtual void v_OnLeftColChange (
			int oldcol,
			int newcol
			);

		//! This is called in response to a popup menu option.
		//!
		//! The vars row and col are where the right mouse button was clicked,
		//! item is the ID you passed to MGUI::MENU::Add in v_OnMenuStart()
		virtual void v_OnMenuCommand (
			int col,
			int row,
			SECTION section,
			int item
			);

		//! Called when the user right clicks and a popup menu is about to be displayed.  
		//!
		//! @return True to enable the menu, false to prevent the menu.  (Note, 
		//! the menu won't popup if empty either)
		//!
		//! This gives you an emtpy MGUI::MENU to populate.
		//!
		//! If you do not specify a delegate when you add items to the menu, the
		//! default delegate will cause it to call the grid's v_OnMenuCommand().
		virtual bool v_OnMenuStart (
			int col,
			int row,
			SECTION section,
			MGUI::MENU& menu
			);
		
		//! Called when the user left clicks on a cell.
		//! 
		//! The point passed in is coordinates relative to the upper
		//! left of the cell where the click took place.
		//! The rect is the extents of said cell.
		//! NOTE:  If you use the event, set *pbProcessed to true
		//! to prevent the event from propagating.
		//! NOTE:  If you have called EnableMenu with true, the popup
		//! menu will be handled instead of OnRClicked.
		virtual void v_OnRClicked (
			int col,
			int row,
			bool bIsMouseDown,
			LRECT2D* rect,
			LPOINT2D* point,
			bool* pbProcessed,
			UINT32 state		//!< ShiftMask, CtrlMask, etc from the XEvent
			);

		//! Called when the focus shifts to a new row.
		virtual void v_OnRowChange (
			int oldrow,
			int newrow
			);
		
		//! Like OnColumnSized, but for rows (Row resizing not implemented anyway).
		virtual void v_OnRowSized (
			int row,
			int* height
			);

		//! Like OnColumnSizing, but for rows (Row resizing not implemented anyway).
		virtual void v_OnRowSizing (
			int row,
			int* height
			);

		//! Called by SetCell() to let you ammend the changes.
		virtual void v_OnSetCell (
			int col,
			int row,
			MIGRIDCELL* pCell
			);

		//! Called when the grid gains focus or when focus shifts to a different section of the grid.
		virtual void v_OnSetFocus (
			SECTION section
			);

		//! Called when the whole thing is being created (Create()).
		virtual void v_OnSetup (
			);

		//! Called when the side heading is double clicked.
		//!
		//! Note that you can have multiple side headings.  The right-most side heading
		//! is column -1, and the decrease going to the left.
		virtual void v_OnSideHeadingDClicked (
			int col,
			int row,
			LRECT2D* pRect,
			LPOINT2D* pPoint,
			bool* pbProcessed
			);

		//! Called when the side heading is left clicked. 
		//!
		//! Note that you can have multiple side headings.  The right-most side heading
		//! is column -1, and the decrease going to the left.
		virtual void v_OnSideHeadingLClicked (
			int col,
			int row,
			bool bIsMouseDown,
			LRECT2D* pRect,
			LPOINT2D* pPoint,
			bool* pbProcessed
			);

		//! Called when the side heading is left clicked.
		//!
		//! Note that you can have multiple side headings.  The right-most side heading
		//! is column -1, and the decrease going to the left
		//! Note, that if menus are enabled, this gets called before the
		//! menu is processed and can disable it by setting *bpProcessed to true.
		virtual void v_OnSideHeadingRClicked (
			int col,
			int row,
			bool bIsMouseDown,
			LRECT2D* pRect,
			LPOINT2D* pPoint,
			bool* pbProcessed
			);

		//! Called when the Grid control is resized.
		//!
		//! Size returned includes headings and scroll bars.
		virtual void v_OnSize (
			int ntype,							//!< Ignore - For compatibility with MFC
			int width,
			int height
			);
	

		//! Called when the top heading is double clicked. 
		//!
		//! Note that you can have multiple side headings.  The right-most side heading
		//! is column -1, and the decrease going to the left.
		virtual void v_OnTopHeadingDClicked (
			int col,
			int row,
			LRECT2D* pRect,
			LPOINT2D* pPoint,
			bool* pbProcessed
			);

		//! Called when the side heading is left clicked. 
		//! 
		//! Note that you can have multiple side headings.  The right-most side heading
		//! is column -1, and the decrease going to the left.
		virtual void v_OnTopHeadingLClicked (
			int col,
			int row,
			bool bIsMouseDown,
			LRECT2D* pRect,
			LPOINT2D* pPoint,
			bool* pbProcessed
			);

		//! Called when the side heading is left clicked. 
		//!
		//! Note that you can have multiple side headings.  The right-most side heading
		//! is column -1, and the decrease going to the left.
		//! Note, that if menus are enabled, this gets called before the
		//! menu is processed and can disable it by setting *bpProcessed to true.
		virtual void v_OnTopHeadingRClicked (
			int col,
			int row,
			bool bIsMouseDown,
			LRECT2D* pRect,
			LPOINT2D* pPoint,
			bool* pbProcessed
			);

		//! Called while vertical scrolling to get the tooltip to show.
		//!
		//! NOTE:  You have to call UseVScrollHints() to enable this.
		virtual bool v_OnVScrollHint (
			int row,
			MISTRING&
			);
			
		//! Redraw everything.
		int RedrawAll (
			);

		//! Redraw the cell.
		int RedrawCell (
			int col,
			int row
			);

		//! Redraw the column.
		int RedrawCol (
			int col
			);
			
		//! Redraw the row.
		int RedrawRow (
			int row
			);

		//! Select a given cell.
		int Select (
			int col, 		//!< Pass -1 to select whole row
			int row
			);

		//! Unselect a given cell.
		int Unselect (
			int col, 		//!< Can pass -1 to unselect whole row
			int row
			);

		//! Select a block of cells.
		int SelectRange (
			int startcol,
			int startrow,
			int endcol,
			int endrow
			);

		//! Set the height of raised borders in pixels.
		int Set3DHeight (
			int height
			);

		//! Sets the speed of row-by-row/col-by-col scrolling when balistic scrolling is enabled.  
		//!
		//! When the mouse is dragged within approximately one inch outside the grid's boundaries, 
		//! scrolling will occur one row/col at a time.  The delay between scrolling
		//! updates is determined by this function (I may not implement this).
		int SetBallisticDelay (
			int millsec
			);

		//! Set the properties for a cell.
		int SetCell (
			int col,
			int row,
			const MIGRIDCELL& pCell
			);

		//! Set the default properties for a column.
		int SetColDefault (
			int col,
			const MIGRIDCELL& Cell
			);

		//! This function sets the _visual_ order of columns.
		int SetColTranslation (
			int VisualColumnNumber,
			int RealColumnNumber
			);

		//! Set the width of a column in pixels.
		int SetColWidth (
			int col,
			int width
			);

		//! Set the default width for new columns.
		int SetDefColWidth (
			int width
			);

		//! Set the default height for new rows.
		//! If not set, the default row height is big enough to hold the tallest
		//! font in the default font list. (Uses MxGetMaxFontHeight() to determine)
		int SetDefRowHeight (
			int height
			);

		//! Set the default properties for grid cells.
		int SetGridDefault (
			const MIGRIDCELL& Cell
			);

		//! Set the default properties for heading cells.
		int SetHeadingDefault (
			const MIGRIDCELL& Cell
			);

		//! Set row highlighting row.
		//!
		//! If true, will hightlight, the whole row that has focus.
		int SetHighlightRow (
			bool bWholeRow
			);

		//! Set scroll mode to SCROLLMODE_Normal or SCROLLMODE_Tracking.
		//!
		//! Tracking mode updates while dragging the scrollbar.  Normal mode doesn't.
		int SetHScrollMode (
			SCROLLMODE mode
			);

		//! Set the height of the horizontal scrollbar.
		int SetHSHeight (
			int height
			);

		//! Set an ImageList for the Grid.  
		//! By default the grid will make a copy of the list.  If you
		//! pass false for MakeACopy, the list will NOT be copied.
		//! This should be used only if the list itself is a static global which
		//! will exist for the life of the program and be destroied on exit
		void SetImageList (
			MGUI::IMAGELIST* ImageList,
			bool MakeACopy = true	//!< If true, MIGRID copy the image list.
			);


		//! Set the user data for a given item in the list.
		//! This version assumes integer data.  SetItemDataPtr() assumes
		//! a pointer.  Note one calls the other.  This would be done by just
		//! overloading the functions, but that makes NULL ambiguious and the
		//! Get methods have to have different names because you can't have
		//! two methods that differ only in the return type.
		void SetItemData (
			int row,
			int col,
			PTRUINT data
			) {
			SetItemDataPtr(row, col, (void*)data);
			}

		//! Set the user data for a given item in the list.
		//! This version assumes integer data.  SetItemDataPtr() assumes
		//! a pointer.  Note one calls the other.  This would be done by just
		//! overloading the functions, but that makes NULL ambiguious and the
		//! Get methods have to have different names because you can't have
		//! two methods that differ only in the return type.
		void SetItemData (
			int row,
			PTRUINT data
			) {
			SetItemDataPtr(row, 0, (void*)data);
			}

		//! Set the user data for a given item in the list.
		//! This version assumes pointer data.  SetItemData() assumes
		//! integer.  Note one calls the other.  This would be done by just
		//! overloading the functions, but that makes NULL ambiguious and the
		//! Get methods have to have different names because you can't have
		//! two methods that differ only in the return type.
		void SetItemDataPtr (
			int row,
			int col,
			void* data
			);

		//! Set the user data for a given item in the list.
		//! This version assumes pointer data.  SetItemData() assumes
		//! integer.  Note one calls the other.  This would be done by just
		//! overloading the functions, but that makes NULL ambiguious and the
		//! Get methods have to have different names because you can't have
		//! two methods that differ only in the return type.
		void SetItemDataPtr (
			int row,
			void* data
			) {
			SetItemDataPtr(row, 0, data);
			}

		//! Set the left-most visable column.
		int SetLeftCol (
			int col
			);

		//! Set the margin for cells.
		int SetMargin (
			int pixels
			);

		//! Enable/Disable multi selection mod3.
		int SetMultiSelectMode (
			bool bEnable
			);

		//! Set the number of columns.
		//!
		//! NOTE: Much more efficient if done while there are no rows.
		int SetNumberCols (
			int cols
			);

		//! Set the number of rows.
		int SetNumberRows (
			int rows
			);
		
		//! Set the default properties for a row.
		int SetRowDefault (
			int col,
			const MIGRIDCELL& Cell
			);

		//! Set the row height.
		int SetRowHeight (
			int row,
			int height
			);

		//! Set the width of side heading columns.
		//!
		//! Note that there can be > 1 side heading, and they're numbered with -1 being the
		//! right most and decrease going to the left.
		int SetSideHeadingColWidth (
			int col,
			int width
			);
		
		//! Set the number of "side headings" (useful for showing row numbers, etc).
		//!
		//! Hint: This is more efficient if you do it while there are no rows.
		//! Note, side3 heading columns are accessed as columns -1, -2, ... -n
		int SetSideHeadingNumCols (
			int cols
			);

		//! Set the "side headings" width.
		int SetSideHeadingWidth (
			int width
			);

		//! Sets the size of the main form in pixels.
		//!
		//! Size includes headings and scroll bars.
		void SetSize (
			const LPOINT2D& size
			);


		//! Set the height of top heading rows.
		//!
		//! Note that there can be > 1 top heading, and they're numbered with -1 being the
		//! bottom most and decrease going up.
		//! Note:  The default TopHeadingRowHeight is the same as the default row height
		//! which can be set with SetDefRowHeight()
		int SetTopHeadingRowHeight (
			int row,
			int height
			);
		
		//! Set the number of "top headings" (useful for showing field names, etc).
		//!
		//! Hint: This is more efficient if you do it while there are no columns.
		//! Note, top heading rows are accessed as rows -1, -2, ... -n
		//!   Heading row -1 is the bottom most heading.
		int SetTopHeadingNumRows (
			int rows
			);

		//! Set the height of all the top headings.
		//!
		//! Note that this does not have to be the sum of all the headings, but that's most logical.
		int SetTopHeadingHeight (
			int height
			);

		//! Set the top row.
		int SetTopRow (
			int row
			);

		//! Set the uniform row height.
		int SetUniformRowHeight (
			bool bUniform
			);

		//! Allow/disable row/column resizing.
		//!
		//! Note that row resizing will not be allowed if SetUniformRowHeight(true).
		int SetUserSizingMode (
			bool bAllowResize
			);

		//! Set scroll mode to SCROLLMODE_Normal or SCROLLMODE_Tracking.
		//! 
		//! Tracking mode updates while dragging the scrollbar.  Normal mode doesn't.
		int SetVScrollMode (
			SCROLLMODE mode
			);

		//! Set the width of the vertical scrollbar in pixels.
		int SetVSWidth (
			int width
			);

		//! Sort the data in the grid.
		//! 
		//! Note that for now, this only works if all the CellData is set via SetCell(), and it 
		//! physically  swaps the internal rows.  After that, everything is done using
		//! the rows in their sorted order, not the order you filled them
		//! in with.  If you want to keep track of the original order somehow,
		//! you can use GetParam and SetParam to hang any private data
		//! you want to off of each cell.  
		int SortBy (
			int* pCols,
			int num,
			bool bAscending
			);

		//! Sort the data in the grid.
		int SortBy (
			int col,
			bool bAscending
			) {
			return (SortBy(&col, 1, bAscending));
			}

		//! Start editing the cell with focus.
		int StartEdit (
			);

		//! Start editing the cell with focus and force-feed it the given key.
		int StartEdit (
			int key
			);

		//! Start editing a given cell and force-feed it the given key.
		int StartEdit (
			int col,
			int row,
			int key
			);

		//! Pops up the popup menu (if any).
		//!
		//! Point is where to pop the menu up relative to the drawing area
		//! XXX (This function doesn't work correctly yet.).
		int StartMenu (
			int col,
			int row,
			LPOINT2D* point,
			SECTION section
			);

		//! Swap two rows.
		//! Note, this will have no effect if the list gets its cell data through OnGetCell().
		void SwapRows (
			INT32 row1,
			INT32 row2
			);

		//! Turns on hints.
		int UseHints (
			bool bUseHints = true
			);

		//! Turns on horizontal scroll hints.
		int UseHScrollHints (
			bool bUseHints = true
			);

		//! Turns on vertical scroll hints.
		int UseVScrollHints (
			bool bUseHints = true
			);

#ifdef X_NATIVE
		//! Used by editing classes to get widget to pop stuff up over.
		Widget GetDA (
			int col,
			int row
			) const {
			return GetDrawingAreaData(col, row).m_widget;
			}
#endif

		//! Used by editing classes to get gc to draw with (drawing into Pixmap).
		MGD::CONTEXT* GetGC (
			int col,
			int row
			) const {
			return GetDrawingAreaData(col, row).m_gc;
			}

		//! Used by editing classes to get gc to draw with (drawing into DrawingArea).
		MGD::CONTEXT* GetDAGC (
			int col,
			int row
			) const {
			return GetDrawingAreaData(col, row).m_dagc;
			}

		//! Set gc exclusive.
		void SetGCExclusive (
			MGD::CONTEXT* gc
			) {
			m_GCExclusive = gc;
			}

		//! Used internally to keep focus rectangle from flickering while editing.
		void SetHideFocus (
			bool hide
			) {
			m_bHideFocus = hide;
			};


//////////////////////////////////////////////////////////////////////////////!
private:
	#ifndef GENERATING_DOXYGEN_OUTPUT


	enum DRAGACTION {
		DRAG_None = 0,
		DRAG_ColBorder,
		DRAG_RowBorder,
		DRAG_ColHeading,
		DRAG_RowHeading,
		DRAG_MultiSelect
		};

	static int HeapSortSwap (
		INT32 rec1, 
		INT32 rec2, 
		void *vdata
		);

	static int HeapSortCompare (
		INT32 rec1, 
		INT32 rec2, 
		void *vdata
		);

	//! Copy constructor private and not implemented so nobody can accidently copy us
	MIGRIDCTRL (
		const MIGRIDCTRL&
		);

	//! assignment operator private and not implemented so nobody can accidently copy us
	MIGRIDCTRL& operator=(
		const MIGRIDCTRL&
		);


	class ROWDATA {
	public:
		int m_height;							//!< Height of this row, 0 means use defualt
		int m_ystart;							//!< Starting y position of this row
		MIGRIDCELL* m_pStyle;				//!< pointer to default style for this row (if set)
		MIGRIDCELL** m_pHeading;			//!< array of cells to hold the row heading(s)
		MIGRIDCELL** m_pCellData;
		};

	class COLDATA {
	public:
		int m_width;
		int m_xstart;
		int m_orderVtoR;						//!< Drawing order (Visual to Real translation)
		int m_orderRtoV;						//!< Drawing order (Real to Visual translation)
		MIGRIDCELL** m_pHeading;			//!< array of cells to hold the column heading(s)
		MIGRIDCELL* m_pStyle;				//!< pointer to default style for this row (if set)
		};

	class TOOLTIPDATA : public MGUI::CTRL_TOOLTIP {
		public:
		TOOLTIPDATA();
		~TOOLTIPDATA();

		MIGRIDCTRL *m_pCtrl;
		bool m_bDataTipShowing;
		bool m_bEnabled;
#ifdef X_NATIVE
		Widget m_widget;
#endif
		private:
			virtual bool v_OnShowRequest(int x, int y);
		};

	struct DADATA {
		DADATA();
		~DADATA();

		void CBExpose(MIGRIDCTRL*);
		void CBResize(MIGRIDCTRL*);
		void DrawLineToDA(int x1, int y1, int x2, int y2);
		void DrawRectToDA(const LRECT2D&);
		void EraseDirtyArea();

		MGD::DEVICE_X* m_gdisp;
		MGD::CONTEXT* m_gc;
		MGD::DEVICE_X* m_dagdisp;
		MGD::CONTEXT* m_dagc;
		TOOLTIPDATA m_tip;
		int m_height;
		int m_width;
		bool m_bDirty;
		LRECT2D m_DirtyArea;					//!< Drawing Area area to clear
#ifdef X_NATIVE
		Widget m_widget;
		Display* m_xdisp;						//!< The X Display needed so the destructor can free m_pixmap
		GC m_xgc;
		Pixmap m_pixmap;
#endif
		};

	void OnContextMenuCommand(MGUI::ID, bool);

#ifdef X_NATIVE
	static void CBT_ToolTipPopUp(void *cbdata, XtIntervalId *IntervalID);
	static void CB_ContextMenu(Widget, void*, void*);
	static void CB_Expose(Widget, MIGRIDCTRL*, void*);
	static void CB_Resize(Widget, MIGRIDCTRL*, void*);
	static void CB_HSB(Widget, MIGRIDCTRL*, void*);
	static void CB_VSB(Widget, MIGRIDCTRL*, void*);
	static void EH_MouseDown(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
	static void EH_MouseUp(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
	static void EH_MouseMove(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
	static void EH_KeyPress(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
	static void EH_SBTip(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
	static void EH_DATip(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
	static void EH_Resize(Widget, MIGRIDCTRL*, XEvent*, Boolean*);
#endif

	void BeginUpdate();		//! Don't blit pixmap to screen until EndUpdate()
	void EndUpdate();			//! Blit changes back to screen
	void Blit();				//! Copy pixmap to screen

#ifdef X_NATIVE
	//! Given an x and y, return a row and column number.  Function
	//! returns 0 for success, -1 if no cell at given point
	int GetAbsoluteCellFromPoint (
		Widget widget,		//!< Which widget is x/y relative to
		int x,		//!< X coordinate relative to upper left of widget
		int y,		//!< Y coordinate relative to upper left of widget
		int* ptcol,	//!< column number returned
		int* ptrow  //!< row number returned
		) const;
#endif

	int GetPrevCol (			//! Returns the column number to the left (visually) of the given column
		int col					//!< Returns -1 if none there
		) const;

	int GetNextCol (			//! Returns the column number to the right (visually) of the given column
		int col					//!< Returns -1 if none there
		) const;

	MIGRIDCELLTYPE* GetCellType (
		int col,
		int row
		);	//! Not const. Has to call non-const GetCellIndirect

#ifdef X_NATIVE
	DADATA* GetDrawingAreaData (
		Widget widget
		);
#endif

	DADATA* GetDrawingAreaData (
		int col,
		int row
		);

	const DADATA& GetDrawingAreaData (
		int col,
		int row
		) const;


	//////////////////////////////////////////////////////////////////////////////!
	//! Private Members

	LPOINT2D	m_CurFocus;
	LPOINT2D m_DataSize;						//!< size of all rows/cols in pixels
	LPOINT2D m_UpperLeft;					//!< upper left corner into data currrently visable
	LPOINT2D m_DefaultCellSize;			//!< Size of default columns (x) and rows (y)

	SIMPLE_ARRAY<LPOINT2D> m_SelectedCells;			//!< Keep sorted by row/col (col -1 == whole row)
	int m_nLeftCol;							//!< number of the left most (partially) visable column
	int m_nRightCol;							//!< number of the left most (partially) visable column
	int m_nBottomRow;							//!< number of the top most (partially) visable row
	int m_nTopRow;								//!< number of the top most (partially) visable row
	int m_nRows;								//!< Number of rows
	int m_nCols;								//!< Number of columns
	MGUI::MENU *m_pMenu;			
	SECTION m_ContextMenuSection;
	LPOINT2D m_ContextMenuCell;
	LPOINT2D m_MouseDownPoint;				//!< Position where mouse was last pressed
	LPOINT2D m_MouseLastPoint;				//!< Position where mouse was last moved
	int m_nDrag;								//!< Which row/column border are we dragging
	int m_CurrentCursor;

	ROWDATA* m_pRowData;
	COLDATA* m_pColData;

#ifdef X_NATIVE
	Time m_MouseDownTime;
	int m_MultiClickTime;
#endif

	bool m_bRecomputeSizesH;
	bool m_bRecomputeSizesV;
	bool m_bAllowColSwapping;
	bool m_bAllowContextMenu;
	bool m_bAllowCellResize;
	bool m_bHilightRow;
	bool m_bMultiSelect;
	bool m_bUniformRowHeight;
	int m_nInternalUpdate;					//!< If true, we're drawing into pixmap and will blit later
	int m_nDisableUpdates;
	bool m_bNeedUpdate;						//!< an attempt to update happened while disabled
	bool m_bHideFocus;
	bool m_bInButton1Event;

	int m_nTopHeadings;
	int* m_nTopHeadingHeight;
	int m_nTopHeadingTotalHeight;

	int m_nSideHeadings;
	int* m_nSideHeadingWidth;
	int m_nSideHeadingTotalWidth;
	MIGRIDCELL** m_pSideHeadingStyle;
	MIGRIDCELL** m_pTopHeadingStyle;

	int m_n3Dheight;
	DRAGACTION m_DragAction;

	MIGRIDCELL m_GridDefault;
	MIGRIDCELL m_HeadingDefault;

	MGD::PIXEL m_ShadowPixelLight;
	MGD::PIXEL m_ShadowPixelDark;
	MGD::PIXEL m_FocusRectPixel;
	MGD::PIXEL m_BackgroundPixel;
#ifdef X_NATIVE
	Widget m_wMainForm;						//!< Main Form
	Widget m_wVSB;								//!< Vertical Scroll Bar
	Widget m_wHSB;								//!< Horizontal Scroll Bar
	Widget m_wPopup;							//!< Popup menu
#endif

	MIGRIDEDITHANDLE* m_pEditHandle;
	MIGRIDCELLTYPE** m_pCellTypes;		//!< Array of cell types
	int m_nCellTypes;							//!< Number of data cell in array

	TOOLTIPDATA m_VSBTip;					//!< Tool tip for the vertical scrollbar
	TOOLTIPDATA m_HSBTip;					//!< Tool tip for the vertical scrollbar

	DADATA m_DAGrid;
	DADATA m_DATop;
	DADATA m_DASide;
	MGD::CONTEXT* m_GCExclusive;						//!< If not NULL, draw cell will skip drawing to any gc but this one.
	LRECT2D m_GridClipRect;
	bool m_bClipGrid;

	SCROLLMODE m_VScrollMode;
	SCROLLMODE m_HScrollMode;

	MGUI::IMAGELIST* m_ImageList;
	bool m_bWeOwnImageList;
	MEMPOOL m_MemPool;

	friend struct DADATA;
	#endif // GENERATING_DOXYGEN_OUTPUT
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MIGRIDCTRL::CELLSTATE)
#endif



#endif	// INC_MI32_MIGRID_H
