/**
 * \file mi32/nasaodl.h
 * \brief NASA_ODL class definition.
 *
 *		This class implements a parser for handling text in the
 *		NASA/JPL/PDS Object Description Language (ODL).
 *		The description of this language may be found in:
 *
 *			Planetary Data System Standards Reference
 *			Chapter 12.  Object Description Language Specification and Usage
 *			http://!pds.jpl.nasa.gov
 *
 *		Notes:
 *			1. When parsing, GROUP sections are treated as OBJECT sections as the
 *			above document discourages the use of GROUP.
 *			2. Multidimensional sequences are not yet supported.
 *			3. Pointer statements are treated identically to attribute assignments.
 *			The pointer character '^' will be included in the label and/or value.
 *			4. Modifying and formatting an ODL tree is not presently supported, but
 *			could be implemented without significant difficulty if needed.
 *
 * \if NODOC
 * $Id: nasaodl.h_v 1.4 2011/11/04 13:01:38 mju Exp $
 *
 * $Log: nasaodl.h_v $
 * Revision 1.4  2011/11/04 13:01:38  mju
 * Remove redundant hdr.
 *
 * Revision 1.3  2003/09/15 13:49:56  fileserver!dwilliss
 * Doxygen
 *
 * Revision 1.2  2002/05/20 15:43:00  mju
 * Add GetItemValue/String to OBJECT.
 *
 * Revision 1.1  2002/05/09 17:58:11  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_NASAODL_H
#define	INC_MI32_NASAODL_H

#ifndef INC_MI32_MILIST_H
#include <mi32/milist.h>
#endif

#ifndef INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

//! NASA/JPL Object Description Language tree container.
class NASA_ODL {
	public:

		class ITEM;
		class OBJECT;
		typedef MILIST<ITEM*> ITEMLIST;
		typedef MILIST<OBJECT*> OBJECTLIST;

		//! Constructor.
		NASA_ODL (
			);

		//! Destructor.
		~NASA_ODL (
			);

		//! Find item given dot-separated path.
		//! @return Pointer to ITEM or NULL if item not found.
		const ITEM* FindItem (
			const char *path					//!< Path, name after last '.' (if any) is treated as item name
			) const;

		//! Find object given dot-separated path.
		//! @return Pointer to OBJECT or NULL if object not found.
		const OBJECT* FindObject (
			const char *path					//!< Path with object names separated by '.'
			) const;

		//! Get pointer to root object.
		const OBJECT* GetRootObject (
			) const { return (m_pRoot); }

		//! Parse string into ODL hierarchy.
		//! This will replace the existing hierarchy if any.
		ERRVALUE Parse (
			const char *string				//!< Complete ODL string to parse
			);

	private:
	#ifndef GENERATING_DOXYGEN_OUTPUT
		OBJECT *m_pRoot;

		NASA_ODL (const NASA_ODL&);
		NASA_ODL& operator= (const NASA_ODL&);
	#endif // GENERATING_DOXYGEN_OUTPUT
	};


//! Item, consisting of name and sequence/set of values.
class NASA_ODL::ITEM {
	public:

		//! Construct item with no assigned values.
		explicit ITEM (
			const char *name
			);

		//! Destructor.
		~ITEM (
			);

		//! Get item name.
		const char* GetName (
			) const { return (m_name); }

		//! Get number of values for item.
		int GetNumValues (
			) const { return (m_Values.GetNumItems()); }

		//! Get item value as number (double).
		//! @return Value, 0.0 if could not convert.
		double GetValue (
			int index = 0						//!< Value index starting at 0
			) const;

		//! Get item string.
		//! @return NULL if no string for specified index.
		const char* GetString (
			int index = 0						//!< Value index starting at 0
			) const;

	private:
	#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<char*> m_Values;
		char *m_name;

		ERRVALUE ParseValues (const char *string);

		//! Unimplemented
		ITEM (const ITEM&);
		ITEM& operator= (const ITEM&);
		friend class NASA_ODL::OBJECT;
	#endif // GENERATING_DOXYGEN_OUTPUT
	};


//! Object consisting of name, list of objects, list of items.
class NASA_ODL::OBJECT {
	public:

		//! Construct with specified name.
		explicit OBJECT (
			const char *name
			);

		//! Destructor.
		virtual ~OBJECT (
			);

		//! Find item given dot-separated path.
		const ITEM* FindItem (
			const char *path					//!< Path, last name is treated as item name
			) const;

		//! Find object by dot-separated path name.
		const OBJECT* FindObject (
			const char *path
			) const;

		//! Get object name.
		const char* GetName (
			) const { return (m_name); }

		//! Get list of items contained in this node.
		const ITEMLIST& GetItemList (
			) const { return (m_ItemList); }

		//! Find item and retrieve item string.
		//! @return String pointer or NULL if could not find.
		const char *GetItemString (
			const char *path,					//!< Path with object names separated by '.'
			int index = 0						//!< Value index starting at 0
			) const;

		//! Find item and retrieve item value.
		//! @return 'true' if object was found, 'false' if not
		bool GetItemValue (
			const char *path,					//!< Path with object names separated by '.'
			double& value,						//!< Value returned
			int index = 0						//!< Value index starting at 0
			) const;

		//! Get list of objects contained in this object.
		const OBJECTLIST& GetObjectList (
			) const { return (m_ObjectList); }

	private:
	#ifndef GENERATING_DOXYGEN_OUTPUT
		OBJECTLIST m_ObjectList;
		ITEMLIST m_ItemList;
		char *m_name;

		//! Default constructor.
		OBJECT ();
		//! Find item by name only.
		const ITEM* FindItemByName (const char *name) const;
		//! Find object by name only.
		const OBJECT* FindObjectByName (const char *name) const;
		//! Parse string and fill in object.
		ERRVALUE Parse (const char *string, const char *& next);

		//! Unimplemented
		OBJECT (const OBJECT&);
		OBJECT& operator= (const OBJECT&);
		friend class NASA_ODL;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

#endif	// INC_MI32_NASAODL_H
