/**
 * \file mi32/quantizer.h
 * \brief QUANTIZER class definitions.
 *
 * \if NODOC
 * $Id: quantizer.h_v 1.6 2011/10/11 21:31:08 mju Exp $
 *
 * $Log: quantizer.h_v $
 * Revision 1.6  2011/10/11 21:31:08  mju
 * Fix dox and remove redundant hdr.
 *
 * Revision 1.5  2011/07/18 14:33:29  mju
 * Add calcPrecision.
 *
 * Revision 1.4  2011/07/08 13:19:16  mju
 * Rename calc methods to include 2/3D.
 * Add calcRaw2/3D.
 *
 * Revision 1.3  2011/06/24 16:08:47  mju
 * Add more set methods.
 *
 * Revision 1.2  2011/06/22 20:36:01  mju
 * More docs.
 *
 * Revision 1.1  2011/06/22 20:26:25  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MI32_QUANTIZER_H
#define	INC_MI32_QUANTIZER_H

#ifndef	INC_MI32_POINT_H
#include <mi32/point.h>
#endif

//=====================================================================================================================

//! Quantizer to apply scale/offset values.
//! Retains scales for both directions to avoid floating-point division for maximum efficiency.
class QUANTIZER_VALUES {
	public:

		//! Default constructor, uses scale of 1 and offset of 0.
		QUANTIZER_VALUES (
			): m_ScaleQtoD(1), m_ScaleDtoQ(1), m_OffsetQtoD(0) { }

		//! Construct with scale and offset.
		QUANTIZER_VALUES (
			double ScaleQtoD,				//!< Scale from "quantized" to "dequantized" must not be 0
			double OffsetQtoD				//!< Offset from "quantized" to "dequantized", applied after scaling
			): m_ScaleQtoD(ScaleQtoD), m_ScaleDtoQ(1.0/ScaleQtoD), m_OffsetQtoD(OffsetQtoD) { }

		//! Calculate "dequantized" value from quantized value, applying scale and offset
		double CalcDouble (
			double ValueQ					//!< Quantized (integer) value
			) const { return (ValueQ * m_ScaleQtoD + m_OffsetQtoD); }

		//! Calculate precision (number of decimal places) based on scale.
		INT16 CalcPrecision () const
			{
			if (m_ScaleQtoD == 0) return (0);
			double iscale = floor(1.0 / fabs(m_ScaleQtoD) + .5);
			return (static_cast<INT16>(FAST_CEIL_NEARINT(log(iscale)/log(10.0))));
			}

		//! Calculate "raw" value with quantization scale/offset removed, with no rounding.
		double CalcRaw (
			double ValueD					//!< Dequantized value
			) const { return ((ValueD - m_OffsetQtoD) * m_ScaleDtoQ); }

		//! Get scale from "quantized" to "dequantized".
		double GetScale () const
			{ return (m_ScaleQtoD); }

		//! Get offset from "quantized" to "dequantized", applied after scaling.
		double GetOffset () const
			{ return (m_OffsetQtoD); }

		//! Set scale and offset values.
		void Set (
			double ScaleQtoD,				//!< Scale from "quantized" to "dequantized" must not be 0
			double OffsetQtoD				//!< Offset from "quantized" to "dequantized", applied after scaling
			) { SetScale(ScaleQtoD); SetOffset(OffsetQtoD); }

		//! Set scale from "quantized" to "dequantized".
		void SetScale (
			double ScaleQtoD				//!< Scale from "quantized" to "dequantized" must not be 0
			) { m_ScaleQtoD = ScaleQtoD; m_ScaleDtoQ = 1.0 / ScaleQtoD; }

		//! Set offset.
		void SetOffset (
			double OffsetQtoD				//!< Offset from "quantized" to "dequantized", applied after scaling
			) { m_OffsetQtoD = OffsetQtoD; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_ScaleQtoD;
		double m_ScaleDtoQ;
		double m_OffsetQtoD;
		#endif
	};

//=====================================================================================================================

//! Quantizer for converting between "integer" and "double" with scale and offset.
template<typename _TypeQuantized> class QUANTIZER : public QUANTIZER_VALUES {
	public:

		//! Default constructor, uses scale of 1 and offset of 0.
		QUANTIZER ()
			{ }

		//! Construct with scale and offset.
		QUANTIZER (
			double ScaleQtoD,				//!< Scale from "quantized" to "double" must not be 0
			double OffsetQtoD				//!< Offset from "quantized" to "double", applied after scaling
			): QUANTIZER_VALUES(ScaleQtoD,OffsetQtoD) { }

		//! Calculate quantized integer value from "double" value.
		//! No range check is performed.
		_TypeQuantized CalcInt (
			double ValueD					//!< Dequantized value
			) const { return (static_cast<_TypeQuantized>(FAST_ROUND_EXACT(CalcRaw(ValueD)))); }

	};

//=====================================================================================================================

//! Quantizer for converting between LPOINT2D and DPOINT2D with scales and offsets.
class QUANTIZER_LPOINT2D {
	public:

		//! Default constructor, uses scales of 1 and offsets of 0.
		QUANTIZER_LPOINT2D ()
			{ }

		//! Construct with scales and offsets.
		QUANTIZER_LPOINT2D (
			double ScaleX,					//!< X scale from "quantized" to "double"
			double ScaleY,					//!< Y scale from "quantized" to "double"
			double OffsetX,				//!< X offset from "quantized" to "double"
			double OffsetY					//!< Y offset from "quantized" to "double"
			): m_X(ScaleX,OffsetX), m_Y(ScaleY,OffsetY)
			{ }

		//! Construct with DPOINT2D scales and offsets.
		QUANTIZER_LPOINT2D (
			const DPOINT2D& Scale,		//!< Scale from "quantized" to "double"
			const DPOINT2D& Offset		//!< Offset from "quantized" to "double"
			): m_X(Scale.x,Offset.x), m_Y(Scale.y,Offset.y)
			{ }

		//! Calculate DPOINT2D from quantized LPOINT2D.
		DPOINT2D CalcDouble2D (
			const LPOINT2D& ptQ			//!< Quantized point
			) const { return (DPOINT2D(m_X.CalcDouble(ptQ.x),m_Y.CalcDouble(ptQ.y))); }

		//! Calculate DPOINT2D from quantized LPOINT2D.
		void CalcDouble2D (
			const LPOINT2D& ptQ,			//!< Quantized point
			DPOINT2D& ptD					//!< Calculated values returned
			) const { ptD.x = m_X.CalcDouble(ptQ.x); ptD.y = m_Y.CalcDouble(ptQ.y); }

		//! Calculate quantized LPOINT2D from DPOINT2D.
		//! No range check is performed.
		LPOINT2D CalcInt2D (
			const DPOINT2D& ptD			//!< Dequantized point
			) const { return (LPOINT2D(m_X.CalcInt(ptD.x),m_Y.CalcInt(ptD.y))); }

		//! Calculate quantized LPOINT2D from DPOINT2D.
		//! No range check is performed.
		void CalcInt2D (
			const DPOINT2D& ptD,			//!< Dequantized point
			LPOINT2D& ptQ					//!< Quantized values returned
			) const { ptQ.x = m_X.CalcInt(ptD.x); ptQ.y = m_Y.CalcInt(ptD.y); }

		//! Calculate "raw" (unrounded) point.
		void CalcRaw2D (
			const DPOINT2D& ptD,			//!< Dequantized point
			DPOINT2D& ptRaw				//!< Raw (unrounded) values returned
			) const { ptRaw.x = m_X.CalcRaw(ptD.x); ptRaw.y = m_Y.CalcRaw(ptD.y); }

		//! Get X quantizer.
		const QUANTIZER<INT32>& GetX () const
			{ return (m_X); }

		//! Get Y quantizer.
		const QUANTIZER<INT32>& GetY () const
			{ return (m_Y); }

		//! Set scale and offset values from DPOINT2D.
		void Set2D (
			const DPOINT2D& Scale,
			const DPOINT2D& Offset
			) { m_X.Set(Scale.x,Offset.x); m_Y.Set(Scale.y,Offset.y); }

		//! Set X quantizer.
		void SetX (
			const QUANTIZER_VALUES& QuantX
			) { m_X.Set(QuantX.GetScale(),QuantX.GetOffset()); }

		//! Set X quantizer scale and offset.
		void SetX (
			double ScaleX,					//! X scale
			double OffsetX					//! X offset
			) { m_X.Set(ScaleX,OffsetX); }

		//! Set Y quantizer.
		void SetY (
			const QUANTIZER_VALUES& QuantY
			) { m_Y.Set(QuantY.GetScale(),QuantY.GetOffset()); }

		//! Set Y quantizer scale and offset.
		void SetY (
			double ScaleY,					//! Y scale
			double OffsetY					//! Y offset
			) { m_Y.Set(ScaleY,OffsetY); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		QUANTIZER<INT32> m_X;
		QUANTIZER<INT32> m_Y;
		#endif
	};

//=====================================================================================================================

//! Quantizer for converting between LPOINT3D and DPOINT3D with scales and offsets.
class QUANTIZER_LPOINT3D : public QUANTIZER_LPOINT2D {
	public:

		//! Default constructor, uses scales of 1 and offsets of 0.
		QUANTIZER_LPOINT3D ()
			{ }

		//! Construct with scales and offsets.
		QUANTIZER_LPOINT3D (
			double ScaleX,					//!< X scale from "quantized" to "double"
			double ScaleY,					//!< Y scale from "quantized" to "double"
			double ScaleZ,					//!< Z scale from "quantized" to "double"
			double OffsetX,				//!< X offset from "quantized" to "double"
			double OffsetY,				//!< Y offset from "quantized" to "double"
			double OffsetZ					//!< Z offset from "quantized" to "double"
			): QUANTIZER_LPOINT2D(ScaleX,ScaleY,OffsetX,OffsetY), m_Z(ScaleZ,OffsetZ) { }

		//! Construct with DPOINT3D scales and offsets.
		QUANTIZER_LPOINT3D (
			const DPOINT3D& Scale,		//!< Scale from "quantized" to "double"
			const DPOINT3D& Offset		//!< Offset from "quantized" to "double"
			): QUANTIZER_LPOINT2D(Scale,Offset), m_Z(Scale.z,Offset.z) { }

		//! Calculate DPOINT3D from quantized LPOINT3D.
		DPOINT3D CalcDouble3D (
			const LPOINT3D& ptQ			//!< Quantized point
			) const { return (DPOINT3D(GetX().CalcDouble(ptQ.x),GetY().CalcDouble(ptQ.y),m_Z.CalcDouble(ptQ.z))); }

		//! Calculate DPOINT3D from quantized LPOINT3D.
		void CalcDouble3D (
			const LPOINT3D& ptQ,			//!< Quantized point
			DPOINT3D& ptD					//!< Calculated values returned
			) const { ptD.x = GetX().CalcDouble(ptQ.x); ptD.y = GetY().CalcDouble(ptQ.y); ptD.z = m_Z.CalcDouble(ptQ.z); }

		//! Calculate quantized LPOINT3D from DPOINT3D.
		//! No range check is performed.
		LPOINT3D CalcInt3D (
			const DPOINT3D& ptD			//!< Dequantized point
			) const { return (LPOINT3D(GetX().CalcInt(ptD.x),GetY().CalcInt(ptD.y),m_Z.CalcInt(ptD.z))); }

		//! Calculate quantized LPOINT3D from DPOINT3D.
		//! No range check is performed.
		void CalcInt3D (
			const DPOINT3D& ptD,			//!< Dequantized point
			LPOINT3D& ptQ					//!< Quantized values returned
			) const { ptQ.x = GetX().CalcInt(ptD.x); ptQ.y = GetY().CalcInt(ptD.y); ptQ.z = m_Z.CalcInt(ptD.z); }

		//! Calculate "raw" (unrounded) point.
		void CalcRaw3D (
			const DPOINT3D& ptD,			//!< Dequantized point
			DPOINT3D& ptRaw				//!< Raw (unrounded) values returned
			) const { CalcRaw2D(ptD,ptRaw); ptRaw.z = m_Z.CalcRaw(ptD.z); }

		//! Get Z quantizer.
		const QUANTIZER<INT32>& GetZ () const
			{ return (m_Z); }

		//! Set scale and offset values from DPOINT3D scale/offset.
		void Set3D (
			const DPOINT3D& Scale,
			const DPOINT3D& Offset
			) { Set2D(Scale,Offset); m_Z.Set(Scale.z,Offset.z); }

		//! Set Z quantizer.
		void SetZ (
			const QUANTIZER_VALUES& QuantZ
			) { m_Z.Set(QuantZ.GetScale(),QuantZ.GetOffset()); }

		//! Set Z quantizer scale and offset.
		void SetZ (
			double ScaleZ,					//! Z scale
			double OffsetZ					//! Z offset
			) { m_Z.Set(ScaleZ,OffsetZ); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		QUANTIZER<INT32> m_Z;
		#endif
	};

//=====================================================================================================================

#endif	// INC_MI32_QUANTIZER_H
