/**
 * \file mi32/ringbuffer.h
 * \brief A Ring Buffer (FIFO)
 *
 * \if NODOC
 * $Log: ringbuffer.h_v $
 * Revision 1.5  2008/09/09 21:50:43  mju
 * Fix doxygen warning.
 *
 * Revision 1.4  2005/03/31 16:57:19  fileserver!dwilliss
 * Rename one of our types to MIUNICODE because it conflicted with a Microsoft #define
 *
 * Revision 1.3  2004/10/20 17:56:00  dfriberg
 * added UCStrChr
 *
 * Revision 1.2  2004/01/22 17:27:53  dwilliss
 * Destructor no longer inline
 *
 * Revision 1.1  2004/01/02 22:59:09  dwilliss
 * Initial revision
 *
 * \endif
 */

#ifndef INC_MI32_RINGBUFFER_H
#define INC_MI32_RINGBUFFER_H

#ifndef INC_MI32_MILIST_H
#include <mi32/milist.h>
#endif

//! A Simple First-In-First-Out (FIFO) buffer
//! This is used internally by our SOCKET classes.  It's implemented in
//! misystem, but I haven't decided if I should export it or move this
//! include file to mi32p
class RINGBUFFER {
	public:
		RINGBUFFER() {}
		~RINGBUFFER();

		//! Return the number of bytes currently available in the buffer.
		INT32 GetNumBytes (
			) const;

		//! Append bytes to the end of the buffer.
		INT32 Append (
			const UINT8* data,
			INT32 numbytes
			);

		//! Pull bytes from the front of the buffer.
		//! \param data Buffer to copy bytes into
		//! \param numbytes The number of bytes to consume.  Can be NULL if the
		//!	data will just be ignored.
		//! \return The number of bytes actually consumed
		INT32 GetBytes (
			UINT8* data,
			INT32 numbytes
			);

		//! Peek at the head of the buffer.
		//! Fills in a pointer pointing to the head of the buffer and returns
		//! the number of bytes available.
		//! Note, the number of bytes returned may not be the total number of
		//! bytes available, but it's the number of contiguous bytes that
		//! can be accessed through the returned pointer.
		INT32 Peek (
			const void*& ptr
			);

		//! Peek at the buffer and find the offset to the given character.
		//! 
		//! \return Offset to character or -1 if the character isn't in the buffer
		INT32 StrChr (
			char c
			) const;

		//! Peek at the buffer and find the offset to the given unicode character.
		//! 
		//! \return Offset to character or -1 if the character isn't in the buffer
		INT32 UCStrChr (
			MIUNICODE c
			) const;

	private:
#ifndef GENERATING_DOXYGEN_OUTPUT
		class BLOCK;
		MILIST<BLOCK*> m_blocks;
#endif
	};


#endif // INC_MI32_RINGBUFFER_H
