/**
 * \file rvc/quadtree.h 
 * \brief interface for the RVC::QSEARCHTREE class.
 *
 * \if NODOC
 * $Id: quadtree.h_v 1.34 2011/12/23 13:50:24 mju Exp $
 *
 * $Log: quadtree.h_v $
 * Revision 1.34  2011/12/23 13:50:24  mju
 * Remove redundant hdr.
 *
 * Revision 1.33  2010/04/12 16:16:51  mju
 * Rename objtype_QuadSearchTree_...
 *
 * Revision 1.32  2010/04/09 21:30:15  mju
 * Add search method using unsignedset.
 *
 * Revision 1.30  2006/10/17 22:48:29  scowan
 * Added validate method.
 *
 * Revision 1.28  2006/06/28 15:41:11  scowan
 * Changed make to take a non-const objitem.
 *
 * Revision 1.26  2006/01/04 21:59:36  scowan
 * Added method to clear and reset the tree.
 *
 * Revision 1.23  2005/12/09 17:29:48  scowan
 * Updated to use credentials.
 *
 * Revision 1.13  2002/01/15 15:56:29  scowan
 * Added set element type to filter class.
 *
 * Revision 1.1  2000/11/28 22:07:16  scowan
 * Initial revision
 * \endif
**/

#ifndef INC_RVC_QUADTREE_H
#define INC_RVC_QUADTREE_H

#ifndef INC_RVC_OBJECT_H
	#include <rvc/object.h>
#endif

#ifndef INC_MI32_ELEMTYPE_H
	#include <mi32/elemtype.h>
#endif

#ifndef	INC_MI32_RECT_H
	#include <mi32/rect.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
class UNSIGNEDSET;
namespace RVC {
	class DESCRIPTOR;
	class GEOMETRIC;
	class QUADTREEHANDLE;
	}
#endif


namespace RVC {

//! Manages RVC::OBJTYPE_QuadSearchTree... objects.  
//!	
//!	RVC Object Types that this class supports: RVC::OBJTYPE_QuadSearchTree_Legacy, RVC::OBJTYPE_QuadSearchTree_List, RVC::OBJTYPE_QuadSearchTree_Set.
//!
//!	Default RVC Object Type for this class: RVC::OBJTYPE_QuadSearchTree_List.
//!
//!	Valid parents for this class:
//!		- RVC::OBJTYPE_Vector
//!		- RVC::OBJTYPE_CAD
//!		- RVC::OBJTYPE_TIN
//!		- RVC::OBJTYPE_SHAPE
class QSEARCHTREE : public RVC::OBJECT {
	public:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		class MAKEPARMS;
		class FILTER_EXISTING;
		class FILTER_EXISTING_CURRENT;
		class FILTER_EXISTING_LEGACY;
		class FILTER_VALID;
		#endif // GENERATING_DOXYGEN_OUTPUT

		//! Fill in the DESCRIPTOR with defaults based on the ELEMTYPE
		static void GenerateDftDescriptor (
			ELEMTYPE ElemType,
			OBJTYPE ParentObjType,
			DESCRIPTOR& Descriptor,
			const char* Source = 0
			);

		//! Default Constructor
		QSEARCHTREE ();

		//! Copy Constructor
		QSEARCHTREE (
			const RVC::QSEARCHTREE& rhs
			);

		//! Destructor
		virtual ~QSEARCHTREE ();

		//! Assignment
		RVC::QSEARCHTREE& operator= (
			const RVC::QSEARCHTREE& rhs
			);

		//! Add an element to the search tree.
		ERRVALUE AddElement (
			INT32 ElemNum,						//!< Element number to add to search tree.
			const DRECT2D& Extents				//!< Extents of element to add
			);

		//! Add an element to the search tree.
		ERRVALUE AddElement (
			INT32 ElemNum,						//!< Element number to add to search tree.
			const DPOINT2D& Point			//!< Location of element to add
			);

		//! Change an element from the search tree.
		ERRVALUE ChangeElement (
			INT32 OldElemNum,								//!< Old element number
			INT32 NewElemNum,								//!< New element number
			const DRECT2D& Extents				//!< Extents of element to change
			);

		//! Change an element from the search tree.
		ERRVALUE ChangeElement (
			INT32 OldElemNum,								//!< Old element number
			INT32 NewElemNum,								//!< New element number
			const DPOINT2D& Point					//!< Location of element to change
			);

		//! Clear all elements out of quadtree object and make a new initial quadtree setup
		ERRVALUE Clear (
			const MAKEPARMS& MakeParms
			);
			
		ERRVALUE CreateFromLegacy (
			QSEARCHTREE& DestObj
			);
			
		//! Remove an element from the search tree.
		ERRVALUE DeleteElement (
			INT32 ElemNum,								//!< Element to remove
			const DRECT2D& Extents				//!< Extents of element to remove
			);

		//! Remove an element from the search tree.
		ERRVALUE DeleteElement (
			INT32 ElemNum,								//!< Element to remove
			const DPOINT2D& Point					//!< Location of element to remove
			);

		//! Obtain the ELEMTYPE associated with this search tree.
		ELEMTYPE GetElemType (
			) const { return (m_ElemType); } 

		//! Make a search tree object.
		ERRVALUE Make (
			const RVC::OBJECT& parent,
			const RVC::DESCRIPTOR& Descriptor,
			const MAKEPARMS& MakeParms,
			const CREDENTIALS *pCredentials RVC_DEFAULT_CREDENTIALS				//!< User credentials passed in, NULL for default behavior
			);

		//! Make a search tree object.
		ERRVALUE Make (
			RVC::OBJITEM& ObjItem,								//!< Parent object and DESCRIPTOR of object to create, new location RETURNED
			const MAKEPARMS& MakeParms,
#ifdef RVC_CREDENTIALS_MANDATORY
			const char *Source,							//!< Source of new spatial tree object, default is application name
			const CREDENTIALS *pCredentials //!< User credentials passed in, NULL for default behavior
#else
			const char *Source = 0,							//!< Source of new spatial tree object, default is application name
			const CREDENTIALS *pCredentials = 0				//!< User credentials passed in, NULL for default behavior
#endif
			);

		//! Find array of elements that fall within rectangle.
		ERRVALUE Search (
			const DRECT2D& Extents, 					//!< Extents to search with
			SIMPLE_ARRAY<INT32>& ElemArray			//!< Array of element numbers returned.
			) const;

		//! Find set of elements that fall within rectangle.
		ERRVALUE Search (
			const DRECT2D& Extents, 					//!< Extents to search with
			UNSIGNEDSET& ElemSet							//!< Set of element numbers returned
			) const;

		ERRVALUE Validate (
			) const;
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		//! Implementation of RVC::OBJECT virtual methods
		virtual ERRVALUE v_CloseObject ();
		virtual OBJTYPE v_GetDftObjectType () const;
		virtual bool v_IsMultiObject () const;
		virtual bool v_IsObjectOpen () const;
		virtual ERRVALUE v_OpenObject (OPENMODE OpenMode, MDLGPARENT parent, const CREDENTIALS *pCredentials);
		virtual bool v_ValidateObjectType (const OBJTYPE otype) const;

		QUADTREEHANDLE* m_QTreeHandle;
		ELEMTYPE m_ElemType;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class QSEARCHTREE::MAKEPARMS {
	public:
		//! Default Constructor
		MAKEPARMS (
			ELEMTYPE ElemType,					//!< Quadtree element type
			const DRECT2D& Extents,
			INT32 InitialNumItems
			) :
			m_ElemType(ElemType),
			m_Extents(Extents),
			m_NumItems(InitialNumItems)
			{}

		ELEMTYPE GetElemType (
			) const { return (m_ElemType); }
	
		const DRECT2D& GetExtents (
			) const { return (m_Extents); }
	
		INT32 GetNumItems (
			) const { return (m_NumItems); }
	
		void SetElemType (
			ELEMTYPE ElemType
			) { m_ElemType = ElemType; }
	
		void SetExtents (
			const DRECT2D& Extents
			) { m_Extents  = Extents; }
	
		void SetNumItems (
			INT32 NumItems
			) { m_NumItems = NumItems; }
	
	private:
		ELEMTYPE m_ElemType;
		DRECT2D m_Extents;
		INT32 m_NumItems;
	};

//! Filter to test for existing QSEARCHTREE object of specific OBJTYPE associated with specific ELEMTYPE.
//! Does not check object timestamp.
class QSEARCHTREE::FILTER_EXISTING : public RVC::OBJECT::FILTER {
	public:

		//! Constructor.
		FILTER_EXISTING (
			OBJTYPE ObjTypeSearchTree,			//!< Type of search tree, OBJTYPE_QuadSearchTree_Legacy, OBJTYPE_QuadSearchTree_List or OBJTYPE_QuadSearchTree_Set
			ELEMTYPE ElemType						//!< Spatial element type tree is associated with
			) : m_ObjTypeSearchTree(ObjTypeSearchTree), m_ElemType(ElemType) {}

		//! Set element type to test for.
		void SetElemType (
			ELEMTYPE ElemType
			) { m_ElemType = ElemType; }
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		OBJTYPE m_ObjTypeSearchTree;
		ELEMTYPE m_ElemType;

		virtual bool Filter (const RVC::OBJECT& obj);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};
	
//! Filter to test for existing current QSEARCHTREE object.
class QSEARCHTREE::FILTER_EXISTING_CURRENT : public QSEARCHTREE::FILTER_EXISTING {
	public:

		//! Construct for specified element type.
		explicit FILTER_EXISTING_CURRENT (
			ELEMTYPE ElemType
			) : FILTER_EXISTING(OBJTYPE_QuadSearchTree_List,ElemType) { }

	};
	
//! Filter to test for "valid" QSEARCHTREE object.
//! To be "valid", must have "modified" timestamp later than parent.
//! If more than one "current" QuadSearchTree_X object exists, only the latest type will be considered to be valid.
class QSEARCHTREE::FILTER_VALID : public RVC::OBJECT::FILTER {
	public:
		FILTER_VALID (
			ELEMTYPE ElemType
			) : m_ElemType(ElemType) { }

		//! Set the element type to test for in searching for the correct quad tree object
		void SetElemType (
			ELEMTYPE ElemType
			) { m_ElemType = ElemType; }
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		ELEMTYPE m_ElemType;
		
		virtual bool Filter (const RVC::OBJECT& obj);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};
	
}		// End of RVC namespace

#endif 	// INC_RVC_QSEARCHTREE_H
