/**
 * \file rvc/uidlgdbfieldvalue.h
 * \brief RVC::UI::DLG_DB_FIELD_VALUE definitions
 *
 * \if NODOC
 * $Id: uidlgdbfieldvalue.h_v 1.4 2010/02/22 18:52:13 mju Exp $
 *
 * $Log: uidlgdbfieldvalue.h_v $
 * Revision 1.4  2010/02/22 18:52:13  mju
 * Specify table/field FOR when create.
 *
 * Revision 1.3  2010/02/22 15:10:29  mju
 * Add settings::save/load.
 *
 * Revision 1.2  2008/09/25 15:10:54  mju
 * Move common methods here and add v_setup and v_updateSettings.
 *
 * Revision 1.1  2008/09/22 19:00:23  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_RVC_UIDLGDBFIELDVALUE_H
#define	INC_RVC_UIDLGDBFIELDVALUE_H

#ifndef	INC_MGUI_DLGSHELL_H
#include <mgui/dlgshell.h>
#endif

#ifndef	INC_RVC_DBTABLE_H
#include <rvc/dbtable.h>
#endif

namespace RVC {
namespace UI {
//=====================================================================================================================

//! Base for database field value view/edit dialogs.
class DLG_DB_FIELD_VALUE : public MGUI::DLGSHELL {
	public:

		//! Base for settings.
		class SETTINGS {
			public:
				//! Determine if has been set as "changed".
				bool IsChanged (
					) const { return (m_IsChanged); }
				//! Load values for specified table/field.
				//! Will reset "changed" status.
				//! @return TRUE if found, FALSE if not, < 0 if error.
				int Load (
					const RVC::DBTABLE& TableFor,
					INT32 FieldNumFor
					);
				//! Reset "changed" status.
				void ResetChanged (
					) { m_IsChanged = false; }
				//! Save values for specified table/field.
				//! Will reset "changed" status.
				ERRVALUE Save (
					const RVC::DBTABLE& TableFor,
					INT32 FieldNumFor
					);
			protected:
				//! Default constructor.
				SETTINGS (
					): m_IsChanged(false) { }
				virtual ~SETTINGS () = 0;
				void SetChanged (
					) { m_IsChanged = true; }
			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				bool m_IsChanged;
				#endif
				//! Get base for object name to store settings in.
				virtual const char* v_GetObjNameBase () const = 0;
				virtual const SERIALIZERITEM* v_SerialGetItemDef () const = 0;
			};

		//! Constructor.
		DLG_DB_FIELD_VALUE (
			const MISTRING& title,
			SETTINGS& settings,
			HELPID helpid = HELPID__None,
			DLGSHELL::STYLE dlgstyle = STYLE_Default
			);

		//! Destructor.
		virtual ~DLG_DB_FIELD_VALUE ();

		//! Get whether allowed to edit.
		bool GetAllowEdit (
			) const { return (m_AllowEdit); }

		//! Get field number.
		INT32 GetFieldNumFor (
			) const { return (m_FieldNumFor); }

		//! Get table value is for.
		const RVC::DBTABLE& GetTableFor (
			) const { return (m_TableFor); }

		//! Get value, placing in appropriate field in record.
		ERRVALUE GetValue (
			RVC::DBTABLE::RECORD& record			//!< Record to place value into
			) { return (v_GetValue(record)); }

		//! Set delegate to call when user explicitly closes dialog.
		void SetDelegateOnUserClose (
			DELEGATE_VOID_NOPARMS delegate
			) { m_DelegateOnUserClose = delegate; }

		//! Set delegate to call when value changed.
		//! Delegate will be passed field number as parameter.
		void SetDelegateOnValueChanged (
			DELEGATE_VOID_INT32 delegate
			) { m_DelegateOnValueChanged = delegate; }

		//! Setup with table and field to view/edit values for.
		//! Must be done only once and before dialog is created.
		ERRVALUE Setup (
			const RVC::DBTABLE& ObjTable,			//!< Table view/edit field value for
			INT32 FieldNum,							//!< Field to view/edit
			bool AllowEdit
			);

		//! Set value from appropriate field in record.
		ERRVALUE SetValue (
			const RVC::DBTABLE::RECORD& record	//!< Record to set value from
			) { return (v_SetValue(record)); }

	protected:

		//! Notify via delegate that value has changed.
		void NotifyValueChanged ();
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		DELEGATE_VOID_NOPARMS m_DelegateOnUserClose;
		DELEGATE_VOID_INT32 m_DelegateOnValueChanged;

		RVC::DBTABLE m_TableFor;
		INT32 m_FieldNumFor;
		bool m_AllowEdit;
		SETTINGS *m_pSettings;

		void SaveSettings ();

		// SHELL/DLGSHELL overrides.
		virtual void v_OnClose ();
		virtual void v_OnCloseRequest ();

		#endif	// doxygen

		// Overridables.

		//! Get value, placing in appropriate field in record.
		virtual ERRVALUE v_GetValue (
			RVC::DBTABLE::RECORD& record			//!< Record to place value into
			) = 0;

		//! Setup with table to view/edit values for.
		virtual ERRVALUE v_Setup (
			const RVC::DBTABLE& ObjTable			//!< Table view/edit field value for
			) = 0;
			
		//! Set value from appropriate field in record.
		virtual ERRVALUE v_SetValue (
			const RVC::DBTABLE::RECORD& record	//!< Record to set value from
			) = 0;

		//! Update settings to current state.
		virtual void v_UpdateSettings (
			) = 0;
			
	};

//=====================================================================================================================
}	// End namespace UI
}	// End namespace RVC

#endif	// INC_RVC_UIDLGDBFIELDVALUE_H

