/**
 * \file imagefiltercacheimage.h <spatmod/imagefiltercacheimage.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_CACHEIMAGE definition.
 *
 * \if NODOC
 * $Id: imagefiltercacheimage.h_v 1.2 2006/11/01 23:11:04 mju Exp $
 *
 * $Log: imagefiltercacheimage.h_v $
 * Revision 1.2  2006/11/01 23:11:04  mju
 * queryIterator.
 *
 * Revision 1.1  2006/07/27 21:49:04  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERCACHEIMAGE_H
#define	INC_SPATMOD_IMAGEFILTERCACHEIMAGE_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Filter to cache entire image.
//! Will cache in memory if below memory threshold, or in temporary file if above threshold.
//! Cache supports both line/strip and tile iteration sequences efficiently.
class FILTER_CACHEIMAGE : public FILTER {
	public:

		//! Resample using specified dimensions and georeference.
		//! The decision whether to actually use memory is done during the Initialize() step.
		FILTER_CACHEIMAGE (
			STAGE *pPreviousStage,					//!< Pointer to previous stage
			bool TakeOwnership,						//!< Filter takes ownership of previous stage and will delete when needed
			double MemoryMaxUsage = 4.0,			//!< Memory usage threshold (0 = none, < .5 = fraction of available, > .5 = MB)
			UINT32 TileRows = 0,						//!< Preferred tile rows, 0 to match input
			UINT32 TileColumns = 0					//!< Preferred tile columns, 0 to match input
			);

		// Destructor.
		virtual ~FILTER_CACHEIMAGE ();

		//! Set whether to retain cached image after PullEnd.
		//! Normally the cache is discarded when PullEnd is done to avoid retaining resources
		//! that are no longer needed.  However, for some pipelines there may be multiple
		//! PullBegin/PullEnd sequences on the same stage, in which case this method should be
		//! used to avoid repeated reloading of the cache.
		void SetKeepAfterPullEnd (
			bool KeepAfterPullEnd
			) { m_KeepAfterPullEnd = KeepAfterPullEnd; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		IMAGE::STAGE *m_pCache;
		double m_MemoryMaxUsage;
		bool m_UseMemory;
		bool m_KeepAfterPullEnd;

		ERRVALUE LoadToMemory ();
		ERRVALUE LoadToTempFile ();

		// BASE/STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullBegin (const SECTION_ITERATOR& iterator);
		virtual ERRVALUE v_PullEnd ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);
		virtual float v_QueryIterator (const SECTION_ITERATOR& iterator) const;

		// Unimplemented
		FILTER_CACHEIMAGE (const FILTER_CACHEIMAGE&);
		FILTER_CACHEIMAGE& operator= (const FILTER_CACHEIMAGE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERCACHEIMAGE_H

