/**
 * \file spatmod/imagefilterlookup.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_LOOKUP definition.
 *
 * \if NODOC
 *
 * $Id: imagefilterlookup.h_v 1.3 2008/08/28 13:56:50 mju Exp $
 *
 * $Log: imagefilterlookup.h_v $
 * Revision 1.3  2008/08/28 13:56:50  mju
 * Fix doxygen file ref.
 *
 * Revision 1.2  2006/11/16 14:26:41  mju
 * Add setContrastParm.
 *
 * Revision 1.1  2006/01/30 16:43:28  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERLOOKUP_H
#define	INC_SPATMOD_IMAGEFILTERLOOKUP_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

class CONTRASTPARM;

namespace SPATMOD {
namespace IMAGE {

class SAMPLETABLE;

//=====================================================================================================================

//! Filter to apply lookup tables to samples in image.
class FILTER_LOOKUP : public FILTER {
	public:

		//! Constructor.
		FILTER_LOOKUP (
			STAGE *pPreviousStage,					//!< Pointer to previous stage
			bool TakeOwnership						//!< Filter takes ownership of previous stage and will delete when needed
			);

		//! Destructor.
		virtual ~FILTER_LOOKUP ();

		//! Get pixel value for samples above maximum in table.
		//! @return Pointer to pixel value, NULL if none set.
		const PIXEL* GetPixelAboveMax (
			) const { return (m_pPixelAboveMax); }

		//! Get pixel value for samples below minimum in table.
		//! @return Pointer to pixel value, NULL if none set.
		const PIXEL* GetPixelBelowMin (
			) const { return (m_pPixelBelowMin); }

		//! Get sample table.
		const SAMPLETABLE& GetSampleTable (
			UINT16 SampleNum							//!< Sample index
			) const;

		//! Set all table and pixel values from CONTRASTPARM.
		//! Resulting sample DATATYPE will be UINT8.
		ERRVALUE SetContrastParm (
			const CONTRASTPARM& ContrastParm,	//!< Contrast parameters
			UINT16 SampleNum = 0						//!< Sample number
			);

		//! Set pixel value to use for samples above maximum.
		//! Pixel properties must match image properties.
		//! If not specified will use table maximum.
		ERRVALUE SetPixelAboveMaximum (
			const PIXEL& PixelAboveMax				//!< Pixel value to use
			);

		//! Set pixel value to use for samples below minimum.
		//! Pixel properties must match image properties.
		//! If not specified will use table minimum.
		ERRVALUE SetPixelBelowMinimum (
			const PIXEL& PixelBelowMin				//!< Pixel value to use
			);

		//! Set sample table.
		//! Any component sample for which a table is not provided will be passed through filter unchanged.
		ERRVALUE SetSampleTable (
			UINT16 SampleNum,							//!< Sample index
			const SAMPLETABLE& SampleTable,		//!< Sample table
			double MinValue = 0,						//!< Minimum data value, corresponds to index 0 in sample table
			double Interval = 1						//!< Interval between data values for computing table index
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SAMPLETABLE *m_SampleTableArray;
		PIXEL *m_pPixelBelowMin;
		PIXEL *m_pPixelAboveMax;
		double *m_SampleMinValueArray;
		double *m_SampleIntervalArray;

		// STAGE overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);
		virtual ERRVALUE v_Initialize ();

		// Unimplemented
		FILTER_LOOKUP (const FILTER_LOOKUP&);
		FILTER_LOOKUP& operator= (const FILTER_LOOKUP&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERLOOKUP_H

